"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

pyproject.toml read table sections

..

.. seealso::

   coverage/inorout.py:523: CoverageWarning: Module logging_strict was previously
   imported, but not measured (module-not-measured)
   https://stackoverflow.com/a/18104544

"""

import sys
import tempfile
import unittest
from pathlib import (
    Path,
    PurePath,
)
from unittest.mock import patch

from drain_swamp.pep518_read import (
    _is_ok,
    find_project_root,
    find_pyproject_toml,
)

if sys.version_info >= (3, 9):  # pragma: no cover
    from collections.abc import Sequence  # noqa: F401 Used by sphinx
else:  # pragma: no cover
    from typing import Sequence  # noqa: F401 Used by sphinx


class Pep518Sections(unittest.TestCase):
    def setUp(self):
        if "__pycache__" in __file__:
            # cached
            self.path_tests = Path(__file__).parent.parent
        else:
            # not cached
            self.path_tests = Path(__file__).parent
        self.cwd = self.path_tests.parent

    def test_is_ok(self):
        """Is not None and a non-empty string

        Vendered from package/module, logging-strict.util.check_type
        """
        invalids = (
            None,  # not str
            "",  # empty string
            0.123,  # not str
            "    ",  # contains only whitespace
        )
        for invalid in invalids:
            out_actual = _is_ok(invalid)
            self.assertFalse(out_actual)

        valids = ("Hello World!",)  # non-empty string
        for valid in valids:
            out_actual = _is_ok(valid)
            self.assertTrue(out_actual)

    def test_find_project_root(self):
        """Check possibilities: .git, .hg, pyproject.toml, or file system root"""
        # "pyproject.toml"
        with (
            tempfile.TemporaryDirectory() as f_d,
            patch("pathlib.Path.cwd", return_value=Path(f_d)),
        ):
            tests = ((f_d, "pyproject.toml"),)
            for t_valid_dirs in tests:
                self.assertIsInstance(t_valid_dirs, tuple)
                valid_dir, reason_expected = t_valid_dirs
                path_dir = Path(valid_dir)
                # create an empty pyproject.toml
                # Do not mkdir .git or .hg
                path_f = path_dir.joinpath("pyproject.toml")
                path_f.touch(mode=0o666, exist_ok=False)

                srcs = (valid_dir,)
                path_project_folder, reason = find_project_root(srcs)
                self.assertTrue(issubclass(type(path_project_folder), PurePath))
                self.assertIsInstance(reason, str)
                self.assertEqual(reason, reason_expected)

                is_found = False
                for child in path_project_folder.iterdir():
                    if child.name == "pyproject.toml":
                        is_found = True
                self.assertTrue(is_found)

        # cwd contains a .git folder, making it ill-suited to test as an empty folder
        # Do not make .git or .hg folder. Do not make pyproject.toml file
        with (
            tempfile.TemporaryDirectory() as f_d,
            patch("pathlib.Path.cwd", return_value=Path(f_d)),
        ):
            tests = (
                ((f_d,), "file system root"),
                ((None,), "file system root"),  # means should use cwd
                (None, "file system root"),  # forgot; should be a Sequence
                (
                    (
                        0.1234,
                        None,
                        0.4321,
                        Path("pyproject.toml"),
                    ),
                    "file system root",
                ),  # filter out non-str
            )
            for t_valid_dirs in tests:
                self.assertIsInstance(t_valid_dirs, Sequence)
                srcs, reason_expected = t_valid_dirs
                if srcs is not None:
                    self.assertIsInstance(srcs, Sequence)
                else:
                    self.assertIsNone(srcs)
                self.assertIsInstance(reason_expected, str)

                path_project_folder, reason = find_project_root(srcs)
                self.assertTrue(issubclass(type(path_project_folder), PurePath))
                self.assertIsInstance(reason, str)
                self.assertEqual(reason, reason_expected)

                is_found = False
                for child in path_project_folder.iterdir():
                    if child.name == "pyproject.toml":
                        is_found = True
                self.assertFalse(is_found)

        # stdin_filename
        # .git > pyproject.toml
        with (
            tempfile.TemporaryDirectory() as f_d,
            patch("pathlib.Path.cwd", return_value=Path(f_d)),
        ):
            path_dir = Path(f_d)
            path_dot_git = path_dir.joinpath(".git")
            path_dot_git.mkdir(
                mode=0o777,
                parents=False,
                exist_ok=False,
            )

            # pyproject.toml does not yet exist
            srcs = ("-",)
            stdin_filename = None
            self.assertIsNone(find_pyproject_toml(srcs, stdin_filename))

            path_f = path_dir.joinpath("pyproject.toml")
            path_f.touch(mode=0o666, exist_ok=False)
            str_toml = (
                "[tool.asz.unittest]\n"
                "util/test_pep518_read.py = 14\n\n"
                "[tool.asz.recipe]\n"
                "util/pep518_read = [14]\n\n"
            )
            path_f.write_text(str_toml)

            """pyproject.toml exists, although stdin_filename not supplied

            make coverage says it's None, unittest and running coverage
            on the module says it's a str.

            It shouldn't be None
            """
            # self.assertIsInstance(find_pyproject_toml(srcs, stdin_filename), str)
            pass

            # pyproject.toml exists, stdin_filename supplied
            stdin_filename = str(path_f)
            self.assertIsInstance(find_pyproject_toml(srcs, stdin_filename), str)

            path_project_folder, reason = find_project_root(srcs, stdin_filename)
            self.assertTrue(issubclass(type(path_project_folder), PurePath))
            self.assertIsInstance(reason, str)
            self.assertEqual(reason, ".git directory")

        # Has .git, .hg folders
        tests = (
            (".git", ".git directory"),
            (".hg", ".hg directory"),
        )
        for t_valid_dirs in tests:
            self.assertIsInstance(t_valid_dirs, tuple)
            valid_dir, reason_expected = t_valid_dirs
            with (
                tempfile.TemporaryDirectory() as f_d,
                patch("pathlib.Path.cwd", return_value=Path(f_d)),
            ):
                path_dir = Path(f_d)
                path_dot_vcs = path_dir.joinpath(valid_dir)
                path_dot_vcs.mkdir(
                    mode=0o777,
                    parents=False,
                    exist_ok=False,
                )
                srcs = (f_d,)
                path_project_folder, reason = find_project_root(srcs)
                self.assertTrue(issubclass(type(path_project_folder), PurePath))
                self.assertIsInstance(reason, str)
                self.assertEqual(reason, reason_expected)

        # search fails
        with (
            tempfile.TemporaryDirectory() as f_d,
            patch("pathlib.Path.cwd", return_value=Path(f_d)),
        ):
            tests = ((f_d, "file system root"),)
            for t_valid_dirs in tests:
                self.assertIsInstance(t_valid_dirs, tuple)
                valid_dir, reason_expected = t_valid_dirs
                srcs = (valid_dir,)

                path_project_folder, reason = find_project_root(srcs)
                self.assertTrue(issubclass(type(path_project_folder), PurePath))
                self.assertIsInstance(reason, str)
                self.assertEqual(reason, reason_expected)

                # must be a tuple[str], not tuple[Path]
                stdin_filename = None
                self.assertIsNone(find_pyproject_toml(srcs, stdin_filename))

            # PermissionError, **not** testing a filesystem base folder
            srcs = ("/root",)
            stdin_filename = None
            with self.assertRaises(PermissionError):
                find_project_root(srcs)
            self.assertIsNone(find_pyproject_toml(srcs, stdin_filename))


if __name__ == "__main__":  # pragma: no cover
    """
    .. code-block:: shell

       python -m tests.test_pep518_read

       coverage run --data-file=".coverage-combine-14" \
       -m unittest discover -t. -s tests -p "test_pep518_read*.py" --buffer

       coverage report --include="*pep518_read*" --no-skip-covered \
       --data-file=".coverage-combine-14"

    Ran 1 test in 0.004s
    Branch Coverage: 100%

    Avoid combine since only one RecipePortion

    .. code-block:: shell

       coverage combine --keep .coverage-combine-14

    """
    unittest.main(tb_locals=True)



import logging
from collections.abc import (
    Generator,
    Sequence,
)
from dataclasses import (
    InitVar,
    dataclass,
)
from pathlib import Path
from typing import (
    Any,
    Final,
)

from .backend_abc import BackendType

__all__ = (
    "lock_compile",
    "refresh_links",
    "unlock_compile",
)

_logger: Final[logging.Logger]
is_module_debug: Final[bool]

def is_piptools() -> bool: ...
def _create_symlinks_relative(src: str, dest: str, cwd_path: str) -> None: ...
def _maintain_symlink(path_cwd: Path, abspath_out: Path) -> None: ...
def lock_compile(inst: BackendType) -> Generator[Path, None, None]: ...
def strip_inline_comments(val: str) -> str: ...
@dataclass
class InFile:
    relpath: str
    stem: str
    constraints: set[str] = ...
    requirements: set[str] = ...

    @staticmethod
    def check_path(cwd: Path, path_to_check: Any) -> None: ...
    def abspath(self, path_package_base: Path) -> Path: ...
    @property
    def depth(self) -> int: ...
    def resolve(self, constraint: str, requirements: set[str]) -> None: ...
    def __hash__(self) -> int: ...
    def __eq__(self, right: object) -> bool: ...

@dataclass
class InFiles:
    cwd: Path
    in_files: InitVar[list[Path]]
    _files: set[InFile] = ...
    _zeroes: set[InFile] = ...

    def __post_init__(self, in_files: Sequence[Path]) -> None: ...
    @property
    def files(self) -> Generator[InFile, None, None]: ...
    @files.setter
    def files(self, val: Any) -> None: ...
    @property
    def zeroes(self) -> Generator[InFile, None, None]: ...
    @zeroes.setter
    def zeroes(self, val: Any) -> None: ...
    def in_generic(self, val: Any, set_name: str | None = "files") -> bool: ...
    def in_zeroes(self, val: Any) -> bool: ...
    def __contains__(self, val: Any) -> bool: ...
    def get_by_relpath(
        self,
        relpath: str,
        set_name: str | None = "files",
    ) -> InFile | None: ...
    def move_zeroes(self) -> None: ...
    def resolve_zeroes(self) -> None: ...
    def resolution_loop(self) -> None: ...
    def write(self) -> Generator[Path, None, None]: ...

def unlock_compile(inst: BackendType) -> Generator[Path, None, None]: ...
def refresh_links(inst: BackendType, is_set_lock: bool | None = None) -> None: ...



from collections.abc import Sequence
from pathlib import Path
from typing import Final

entrypoint_name: Final[str]

help_pyproject_toml: Final[str]
help_required: Final[str]
help_optionals: Final[str]
help_extras: Final[str]

def main() -> None: ...
def dependencies_lock(
    path: Path,
    required: Path | None,
    optionals: Sequence[Path] = (),
    extras: Sequence[str] = (),
    additional_folders: Sequence[Path] = (),
    snippet_co: str | None = None,
) -> None: ...
def dependencies_unlock(
    path: Path,
    required: Path | None,
    optionals: Sequence[Path] = (),
    extras: Sequence[str] = (),
    additional_folders: Sequence[Path] = (),
    snippet_co: str | None = None,
) -> None: ...



import copy
import re
import shutil
import sys
from collections.abc import Sequence
from pathlib import (
    Path,
    PurePath,
)
from typing import Any

import pytest

from drain_swamp._run_cmd import run_cmd
from drain_swamp.backend_abc import (
    get_optionals_cli,
    get_optionals_pyproject_toml,
    get_required_pyproject_toml,
)

from .wd_wrapper import WorkDir


class FileRegression:
    """
    .. todo:: when Sphinx<=6 is dropped

       Remove line starting with re.escape(" translation_progress=

    .. todo:: when Sphinx<7.2 is dropped

       Remove line starting with original_url=

    """

    ignores = (
        # Remove when support for Sphinx<=6 is dropped,
        re.escape(" translation_progress=\"{'total': 0, 'translated': 0}\""),
        # Remove when support for Sphinx<7.2 is dropped,
        r"original_uri=\"[^\"]*\"\s",
    )

    def __init__(self, file_regression: "FileRegression") -> None:
        self.file_regression = file_regression

    def check(self, data: str, **kwargs: dict[str, Any]) -> str:
        return self.file_regression.check(self._strip_ignores(data), **kwargs)

    def _strip_ignores(self, data: str) -> str:
        cls = type(self)
        for ig in cls.ignores:
            data = re.sub(ig, "", data)
        return data


@pytest.fixture()
def file_regression(file_regression: "FileRegression") -> FileRegression:
    """Comparison files will need updating.

    .. seealso::

       Awaiting resolution of `pytest-regressions#32 <https://github.com/ESSS/pytest-regressions/issues/32>`_

    """
    return FileRegression(file_regression)


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """
    attach each test's TestReport to the test Item so fixtures can
    decide how to finalize based on the test result.

    fixtures can access the TestReport from the `request` fixture at
    `request.node.test_report`.

    .. seealso::

       https://stackoverflow.com/a/70598731

    """
    test_report = (yield).get_result()
    if test_report.when == "call":
        item.test_report = test_report


@pytest.fixture()
def has_logging_occurred():
    """Display caplog capture text

    Usage

    .. code-block: text

       import pytest
       import logging
       import logging.config
       from drain_swamp.constants import g_app_name, LOGGING

       def test_something(caplog, has_logging_occurred):
           LOGGING['loggers'][g_app_name]['propagate'] = True
           logging.config.dictConfig(LOGGING)
           logger = logging.getLogger(name=g_app_name)
           logger.addHandler(hdlr=caplog.handler)
           caplog.handler.level = logger.level
           assert has_logging_occurred(caplog)

    .. seealso::

       https://github.com/pytest-dev/pytest/discussions/11011
       https://github.com/thebuzzstop/pytest_caplog/tree/master
       `pass params fixture <https://stackoverflow.com/a/44701916>`_

    """

    def _method(caplog) -> bool:
        # log at all logging levels
        # logger was first argument, but removed it
        # import logging
        # logger.debug('DEBUG: log entry captured')
        # logger.info('INFO: log entry captured')
        # logger.error('ERROR: log entry captured')
        # logger.warning('WARNING: log entry captured')
        # display capture log
        print("\nCAPLOG:")
        output = caplog.text.rstrip("\n").split(sep="\n")
        if output == [""]:
            print("Nothing captured")
            return False
        for i in range(len(output)):
            print(f"{i}: {output[i]}")
        return True

    return _method


@pytest.fixture()
def prepare_folders_files(request):
    """Prepare folders and files within folder"""

    set_folders = set()

    def _method(seq_rel_paths, tmp_path):
        set_abs_paths = set()
        is_seq = seq_rel_paths is not None and (
            isinstance(seq_rel_paths, Sequence) or isinstance(seq_rel_paths, set)
        )
        if is_seq:
            for posix in seq_rel_paths:
                if isinstance(posix, str):
                    abs_path = tmp_path.joinpath(*posix.split("/"))
                elif issubclass(type(posix), PurePath):
                    if not posix.is_absolute():
                        abs_path = tmp_path / posix
                    else:  # pragma: no cover
                        # already absolute
                        abs_path = posix
                else:
                    abs_path = None

                if abs_path is not None:
                    set_abs_paths.add(abs_path)
                    set_folders.add(abs_path.parent)
                    abs_path.parent.mkdir(parents=True, exist_ok=True)
                    abs_path.touch()
        else:
            abs_path = None

        return set_abs_paths

    yield _method

    # cleanup
    if request.node.test_report.outcome == "passed":
        for abspath_folder in set_folders:
            shutil.rmtree(abspath_folder, ignore_errors=True)


@pytest.fixture()
def prep_pyproject_toml(request):
    """cli doesn't offer a ``parent_dir`` option to bypass ``--path``.
    Instead copy and rename the test ``pyproject.toml`` to the ``tmp_path``

    :param p_toml_file:

       Path to a ``pyproject.toml``. A copy will be made, original untouched

    :type p_toml_file: pathlib.Path
    :param path_dest_dir: destination tmp_path
    :type path_dest_dir: pathlib.Path
    :returns: Path to the copied and renamed file within it's new home, temp folder
    :rtype: pathlib.Path
    """
    lst_delete_me = []

    def _method(p_toml_file, path_dest_dir, rename="pyproject.toml"):
        if p_toml_file is not None and issubclass(type(p_toml_file), PurePath):
            # copy
            path_dest = path_dest_dir.joinpath(p_toml_file.name)
            shutil.copy(p_toml_file, path_dest_dir)
            # rename
            path_f = path_dest.parent.joinpath(rename)
            shutil.move(path_dest, path_f)
            ret = path_f
            lst_delete_me.append(path_f)
        else:
            ret = None

        return ret

    yield _method

    # cleanup
    if request.node.test_report.outcome == "passed":
        for path_delete_me in lst_delete_me:
            if (
                path_delete_me is not None
                and issubclass(type(path_delete_me), PurePath)
                and path_delete_me.exists()
                and path_delete_me.is_file()
            ):
                path_delete_me.unlink()


@pytest.fixture()
def prep_cmd_unlock_lock():
    """Prepare the cmd for:

    - drain_swamp.cli_unlock.dependencies_lock
    - drain_swamp.cli_unlock.dependencies_unlock

    :param path_tmp_dir: temp folder path
    :type path_tmp_dir: pathlib.Path
    :param t_req:
    :type t_req: tuple[str, pathlib.Path] | None
    :param d_opts: cli optionals target / relative path
    :type d_opts: dict[str, pathlib.Path]
    :param add_folders:

       relative paths to additional folders which contain ``.in`` requirements files

    :type add_folders: list[pathlib.Path]
    :returns:

       cmd usable with :py:mod:`subprocess` or :py:class:`click.testing.CliRunner`

    :rtype: list[str | pathlib.Path]
    """

    def _method(
        path_tmp_dir,
        t_req=(),
        d_opts={},
        add_folders=(),
        snip_co=None,
    ):
        cmd = [
            "--path",
            path_tmp_dir,
        ]

        if t_req is not None and len(t_req) == 2:
            cmd.append("--required")
            cmd.append(t_req[0])
            cmd.append(t_req[1])

        for target, path_rel in d_opts.items():
            cmd.append("--optional")
            cmd.append(target)
            cmd.append(path_rel)

        for path_dir in add_folders:
            cmd.append("--dir")
            cmd.append(path_dir)

        if snip_co is not None:
            cmd.append("--snip")
            cmd.append(snip_co)

        return cmd

    return _method


@pytest.fixture()
def prepare_files_empties(prepare_folders_files):
    """
    :param d_pyproject_toml: dict of a ``pyproject.toml``
    :type d_pyproject_toml: dict[str, typing.Any]
    :param path_dir: A temporary folder to place the file tree
    :type path_dir: pathlib.Path
    :param d_add_files:

       Default empty dict. Additional files. Not to be confused with
       cli optionals. Relative paths

    :type d_add_files: dict[str, pathlib.Path]
    """

    def _method(d_pyproject_toml, path_dir, d_add_files={}, d_optionals={}):
        # pyproject.toml optionals + additional files (w/o optionals_cli)
        d_both = copy.deepcopy(d_add_files)
        get_optionals_pyproject_toml(
            d_both,
            d_pyproject_toml,
            path_dir,
            is_bypass=True,
        )
        seq_prepare_these = list(d_both.values())
        prepare_folders_files(seq_prepare_these, path_dir)

        d_both.clear()

        # optionals_cli
        d_both = copy.deepcopy(d_optionals)
        get_optionals_cli(
            d_both,
            path_dir,
            d_optionals,
        )
        seq_prepare_these = list(d_both.values())
        prepare_folders_files(seq_prepare_these, path_dir)

        # required -- pyproject.toml
        t_required = get_required_pyproject_toml(
            d_pyproject_toml,
            path_dir,
            is_bypass=True,
        )
        if (
            t_required is not None
            and isinstance(t_required, Sequence)
            and len(t_required) == 2
        ):
            seq_prepare_these = (t_required[1],)
            prepare_folders_files(seq_prepare_these, path_dir)

    return _method


@pytest.fixture
def path_project_base():
    """Get the project base folder"""

    def _method():
        if "__pycache__" in __file__:
            # cached
            path_tests = Path(__file__).parent.parent
        else:
            # not cached
            path_tests = Path(__file__).parent
        path_cwd = path_tests.parent

        return path_cwd

    return _method


@pytest.fixture()
def wd(tmp_path: Path) -> WorkDir:
    """https://github.com/pypa/setuptools_scm/blob/main/testing/conftest.py"""
    target_wd = tmp_path.resolve() / "wd"
    target_wd.mkdir()
    return WorkDir(target_wd)


@pytest.fixture
def verify_tag_version():
    """Verify version file contains a given semantic version str

    :param cwd: package base folder
    :type cwd: pathlib.Path
    :param sem_ver_str: expected semantic version
    :type sem_ver_str: str
    :returns: True if versions match otherwise False
    :rtype: bool
    """

    def _method(cwd, sem_ver_str):
        p_bin_dir = Path(sys.executable).parent
        drainswamp_path = str(p_bin_dir.joinpath("drain-swamp"))
        cmd = [
            drainswamp_path,
            "tag",
        ]
        t_ret = run_cmd(cmd, cwd=cwd)
        out, err, exit_code, exc = t_ret
        is_eq = out == sem_ver_str

        return is_eq

    return _method



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_run_cmd.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_run_cmd.py

"""

import pytest

from drain_swamp._run_cmd import run_cmd


def test_run_cmd(tmp_path):
    # pytest --showlocals --log-level INFO -k "test_run_cmd" tests
    # expecting Sequence
    cwd = tmp_path
    invalids = (0.1234, None)
    for invalid in invalids:
        with pytest.raises(TypeError):
            run_cmd(invalid)

        # env is None or not os._Environ
        cmd = ("bin/true",)
        t_ret = run_cmd(cmd, env=invalid, cwd=cwd)
        out, err, exit_code, str_exc = t_ret
        assert str_exc == """No such file or directory bin/true"""

        cmd = ("/bin/true",)
        t_ret = run_cmd(cmd, env=invalid, cwd=cwd)
        out, err, exit_code, str_exc = t_ret
        assert exit_code == 0



import logging
from typing import Any

from drain_swamp.monkey.hooks import markers

logger: logging.Logger

def is_package_installed(app_name: str) -> bool: ...
def _kind(
    config_settings: dict[str, Any] | None, fallback: str | None = "tag"
) -> str: ...
@markers.hook_impl(specname="ds_before_version_infer")  # noqa: Y020
def on_version_infer(config_settings: dict[str, Any]) -> str | None: ...



from collections.abc import Callable
from typing import Any

import setuptools

def version_keyword(
    dist: setuptools.Distribution,
    keyword: str,
    value: bool | dict[str, Any] | Callable[[], dict[str, Any]],
) -> None: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

..

These :pep:`518` (aka pyproject.toml) functions are **not** app specific

These functions are lifted from the black project. With minor changes:

- Removed black specific version handling
- Added sphinx style code documentation
- Removed py38 styles typing like: Tuple or Dict
- Import Sequence correctly

.. seealso::

   ``pyproject.toml`` handling

   - :py:mod:`black`

.. note:: Flake8 config files handling

   Flake8
   `load_config <https://github.com/PyCQA/flake8/blob/fb9a02aaf77b56fcad4320971e7edca0cea93489/src/flake8/options/config.py#L56>`_
   function

   Monkeypatch of configparser to support pyproject.toml,
   `ConfigParserTomlMixin <https://github.com/csachs/pyproject-flake8/blob/16b9dd4d2f19dcf0bfb3a3110f98227627cd67fe/pflake8/__init__.py#L22>`_

   Apply monkeypatch
   `FixFilenames.apply <https://github.com/csachs/pyproject-flake8/blob/16b9dd4d2f19dcf0bfb3a3110f98227627cd67fe/pflake8/__init__.py#L86>`_

**Module private variables**

.. py:attribute:: __all__
   :type: tuple[str]
   :value: ("find_project_root", "find_pyproject_toml")

   Exported objects from this module

**Module objects**

"""

from __future__ import annotations

import sys
from pathlib import Path
from typing import Any

if sys.version_info >= (3, 9):  # pragma: no cover
    from collections.abc import Sequence  # noqa: F401 Used by sphinx
else:  # pragma: no cover
    from typing import Sequence  # noqa: F401 Used by sphinx

__package__ = "drain_swamp"
__all__ = (
    "find_project_root",
    "find_pyproject_toml",
)


def _is_ok(test):
    """Check if non-empty str

    Edge case: contains only whitespace --> ``False``

    :param test: variable to test
    :type test: typing.Any | None
    :returns: ``True`` if non-empty str otherwise ``False``
    :rtype: bool

    .. note::

       Vendored logging-strict.util.check_type.is_ok, so current module
       can be stand alone

    """
    ret = False
    is_str = test is not None and isinstance(test, str)
    if is_str:
        # Edge case: contains only whitespace
        str_stripped = test.strip()
        ret = len(str_stripped) != 0
    else:
        ret = False

    return ret


def find_project_root(srcs, stdin_filename=None):
    """Return folder containing .git, .hg, or ``pyproject.toml``.

    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.

    Returns a two-tuple with the first element as the project root path and
    the second element as a string describing the method by which the
    project root was discovered.

    :param srcs:

       Files or folders, for files will take the parent folder.
       Potential folders that may contain ``pyproject.toml``

    :type srcs: collections.abc.Sequence[typing.Any] | None

    :param stdin_filename:

       Default ``None``. stdin file name, considers files parent as the
       project top folder

    :type stdin_filename: str | None
    :returns:

       Folder containing .git, .hg, or ``pyproject.toml``, will be a common
       parent of all files and directories passed in
       :paramref:`~drain_swamp.pep518_read.find_project_root.params.srcs`

    :rtype: tuple[pathlib.Path, str]
    :raises:

       - :py:exc:`PermissionError` -- Unreadable folder. Ungracefully handled

    .. note::

       Passing ``pyproject.toml``, by stdin, could be useful for testing recipes

    .. warning::

       functools.lru_cache causes optional parameter stdin_filename to **NOT BE OPTIONAL**

    .. seealso::

       :py:func:`black.files.find_project_root` source and credit Black

    """

    def is_sequence_empty(some_sequence: Sequence[Any] | None) -> bool:
        ret = (
            some_sequence is not None
            and isinstance(some_sequence, Sequence)
            and len(some_sequence) == 0
        )
        return ret

    def is_none(arg: Any) -> bool:
        ret = arg is None
        return ret

    def is_sequence_none(some_sequence: Sequence[Any] | None) -> bool:
        ret = (
            some_sequence is not None
            and isinstance(some_sequence, Sequence)
            and len(some_sequence) == 1
            and some_sequence[0] is None
        )
        return ret

    is_use_cwd = is_none(srcs) or is_sequence_none(srcs)
    if is_use_cwd:
        # Intention is: use cwd
        # None or Sequence[None]
        cwd_path = str(Path.cwd().resolve())
        srcs = []
        srcs.append(cwd_path)
    else:
        """Signature is intended to be Sequence[str], but assume
        Sequence[Any]. Filter out non-str, including None and str
        containing only whitespace"""
        srcs = [src for src in srcs if _is_ok(src)]

        if stdin_filename is not None and len(srcs) != 0:
            srcs = list(stdin_filename if s == "-" else s for s in srcs)
        elif stdin_filename is None and len(srcs) != 0:  # pragma: no cover
            # Added "-", but didn't supply :paramref:`stdin_filename`
            srcs = [s for s in srcs if s != "-"]
        else:  # pragma: no cover
            pass

    if is_none(srcs) or is_sequence_empty(srcs):  # pragma: no cover fallback
        srcs = [str(Path.cwd().resolve())]
    else:  # pragma: no cover
        pass

    path_srcs = [Path(Path.cwd(), src).resolve() for src in srcs]

    # A list of lists of parents for each 'src'. 'src' is included as a
    # "parent" of itself if it is a directory
    src_parents = [
        list(path.parents) + ([path] if path.is_dir() else []) for path in path_srcs
    ]

    common_base = max(
        set.intersection(*(set(parents) for parents in src_parents)),
        key=lambda path: path.parts,
    )

    for directory in (common_base, *common_base.parents):
        if (directory / ".git").exists():
            return directory, ".git directory"
        else:  # pragma: no cover
            pass

        if (directory / ".hg").is_dir():
            return directory, ".hg directory"
        else:  # pragma: no cover
            pass

        if (directory / "pyproject.toml").is_file():
            return directory, "pyproject.toml"
        else:  # pragma: no cover
            pass

    return directory, "file system root"


def find_pyproject_toml(path_search_start, stdin_filename):
    """Find the absolute filepath to a ``pyproject.toml`` if it exists

    :param path_search_start:

       absolute paths of files or folders to start search for project base folder

    :type path_search_start: tuple[str, ...]
    :param stdin_filename: ``pyproject.toml`` passed into stdin. May be a file
    :type stdin_filename: str | ``None``
    :returns: Absolute path to project ``pyproject.toml`` otherwise ``None``
    :rtype: str | ``None``
    """
    """2nd item, reason, is string describing the method by which the project
    root was discovered"""
    try:
        path_project_root, _ = find_project_root(path_search_start, stdin_filename)
    except PermissionError:
        return None
    else:
        path_pyproject_toml = path_project_root / "pyproject.toml"
        if path_pyproject_toml.is_file():
            return str(path_pyproject_toml)
        else:
            return None



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Build environment for taking current or tag version

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_versioning_integration.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_versioning_integration.py

"""

import logging
import logging.config
from pathlib import Path

import pytest

from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.version_file.dump_version import write_version_files
from drain_swamp.version_semantic import _tag_version

testdata_version_file_read_normal = (
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "complete-manage-pip-prod-unlock.pyproject_toml",
        ),
        "src/complete_awesome_perfect/_version.py",
        "0.0.5",
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "complete-version-txt.pyproject_toml",
        ),
        "VERSION.txt",
        "0.0.5",
    ),
    pytest.param(
        Path(__file__).parent.joinpath(
            "_bad_files",
            "complete-version-other.pyproject_toml",
        ),
        "VERSION.rst",
        None,
        marks=pytest.mark.xfail(raises=ValueError),
    ),
)
ids_version_file_read_normal = (
    "tool.setuptools.dynamic.version attr key",
    "tool.setuptools.dynamic.version file key",
    "unsupported version file type. rst",
)


@pytest.mark.parametrize(
    "path_config_src, version_file_relpath, kind_expected",
    testdata_version_file_read_normal,
    ids=ids_version_file_read_normal,
)
def test_version_file_read_normal(
    path_config_src,
    version_file_relpath,
    kind_expected,
    tmp_path,
    prep_pyproject_toml,
    prepare_folders_files,
    caplog,
    has_logging_occurred,
):
    """version_file contains the tag version, retrieve it

    Does not depend on git
    """
    # pytest --showlocals --log-level INFO -k "test_version_file_read_normal" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    kind = "0.0.5"

    relpath_version_file = Path(version_file_relpath)
    write_to = relpath_version_file

    # prepare
    #    base package folder
    path_cwd = tmp_path / "complete_awesome_perfect"
    path_cwd.mkdir()

    #    package code folder
    path_dir_dest = path_cwd.joinpath("src", "complete_awesome_perfect")
    path_dir_dest.mkdir(parents=True)

    path_f = path_cwd / relpath_version_file

    #    pyproject.toml
    prep_pyproject_toml(path_config_src, path_cwd)

    # get tag version, but no existing file
    next_versions = (
        None,
        "",
        "   ",  # whitespace only
        0.1234,  # unsupported version
    )
    for next_version in next_versions:
        with pytest.warns(UserWarning) as w:
            # static dependency would be read from pyproject.toml
            # so actual is not always None
            actual = _tag_version(
                next_version=next_version,
                path=path_cwd,
            )
            assert len(w) == 1
            assert str(w[0].message) != 0
            assert actual is None

    #    _version.py|VERSION.txt -- empty file
    seq_empties = (version_file_relpath,)
    prepare_folders_files(seq_empties, path_cwd)

    # act
    write_version_files(kind, path_cwd, write_to, None)

    # confirm
    contents_actual = path_f.read_text()
    if kind_expected is None:
        assert contents_actual is None
    else:
        assert kind in contents_actual

    # next_version empty str | just whitespace | None --> no write to version_file
    for next_version in next_versions:
        actual = _tag_version(
            next_version=next_version,
            path=path_cwd,
        )
        assert isinstance(actual, str)

    # assert has_logging_occurred(caplog)
    pass


testdata_version_file_read_special = (
    (
        Path(__file__).parent.joinpath(
            "_bad_files",
            "static-version.pyproject_toml",
        ),
        "src/complete_awesome_perfect/_version.py",
        "0.0.5",
    ),
)
ids_version_file_read_special = ("project.version, not dynamic",)


@pytest.mark.parametrize(
    "path_config_src, version_file_relpath, kind_expected",
    testdata_version_file_read_special,
    ids=ids_version_file_read_special,
)
def test_version_file_read_special(
    path_config_src,
    version_file_relpath,
    kind_expected,
    tmp_path,
    prep_pyproject_toml,
    prepare_folders_files,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_version_file_read_special" tests
    kind = "0.0.5"
    relpath_version_file = Path(version_file_relpath)
    write_to = relpath_version_file

    # prepare
    #    base package folder
    path_cwd = tmp_path / "complete_awesome_perfect"
    path_cwd.mkdir()

    #    package code folder
    path_dir_dest = path_cwd.joinpath("src", "complete_awesome_perfect")
    path_dir_dest.mkdir(parents=True)

    path_f = path_cwd / relpath_version_file

    #    pyproject.toml
    prep_pyproject_toml(path_config_src, path_cwd)

    #    _version.py|VERSION.txt -- empty file
    seq_empties = (version_file_relpath,)
    prepare_folders_files(seq_empties, path_cwd)

    # act
    write_version_files(kind, path_cwd, write_to, None)

    # confirm
    contents_actual = path_f.read_text()
    if kind_expected is None:
        assert contents_actual is None
    else:
        assert kind in contents_actual

    next_versions = (
        None,
        "",
        "   ",  # whitespace only
        0.1234,  # unsupported version
    )
    for next_version in next_versions:
        actual = _tag_version(
            next_version=next_version,
            path=path_cwd,
        )
        assert isinstance(actual, str)


def test_version_file_read_invalid(
    tmp_path,
    prep_pyproject_toml,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_version_file_read_invalid" tests
    path_config_src = Path(__file__).parent.joinpath(
        "_good_files",
        "backend-only.pyproject_toml",
    )
    # prepare

    #    base package folder
    path_cwd = tmp_path / "complete_awesome_perfect"
    path_cwd.mkdir()

    #    pyproject.toml
    prep_pyproject_toml(path_config_src, path_cwd)

    next_versions = (
        None,
        "",
        "   ",  # whitespace only
        0.1234,  # unsupported version
    )
    for next_version in next_versions:
        with pytest.warns(UserWarning) as w:
            actual = _tag_version(
                next_version=next_version,
                path=path_cwd,
            )
            assert len(w) == 1
            assert str(w[0].message) != 0
            assert actual is None



import abc
import logging
from collections.abc import Sequence
from pathlib import Path
from typing import (
    Any,
    NamedTuple,
)

from setuptools_scm._integration.toml import TOML_RESULT

__all__ = (
    "ReadPyproject",
    "ReadPyprojectStrict",
)

log: logging.Logger

class PyProjectData(NamedTuple):
    path: Path
    tool_name: str
    project: TOML_RESULT
    section: TOML_RESULT

    @property
    def project_name(self) -> str | None: ...

class ReadPyprojectBase(abc.ABC):
    @abc.abstractmethod
    def update(self, d_target: dict[str, Any], d_other: dict[str, Any]) -> None: ...
    def __call__(
        self,
        path: Path = ...,
        tool_name: Sequence[str] | str = ...,
        require_section: bool = True,
    ) -> PyProjectData: ...

class ReadPyproject(ReadPyprojectBase):
    def update(self, d_target: dict[str, Any], d_other: dict[str, Any]) -> None: ...

class ReadPyprojectStrict(ReadPyprojectBase):
    def update(self, d_target: dict[str, Any], d_other: dict[str, Any]) -> None: ...

def read_pyproject(
    path: Path = ...,
    tool_name: Sequence[str] | str = ...,
    require_section: bool = True,
) -> PyProjectData: ...



from collections.abc import (
    Iterator,
    Sequence,
)
from pathlib import Path
from typing import (
    Any,
    ClassVar,
)

from .backend_abc import BackendType

__all__ = ("BackendSetupTools",)

class BackendSetupTools(BackendType):
    _backend: str
    _extras: tuple[str, ...]
    _path_required: tuple[str, Path] | None
    _paths_optional: dict[str, Path]

    BACKEND_NAME: ClassVar[str] = "setuptools"
    PYPROJECT_TOML_SECTION_NAME: ClassVar[str] = "tool.setuptools.dynamic"
    __slots__ = ("path_config", "_extras", "_path_required", "_paths_optional")

    def __init__(
        self,
        d_pyproject_toml: dict[str, Any],
        path_config: Path,
        required: Path | None = None,
        optionals: Sequence[Path] = (),
        extras: Sequence[str] = (),
        parent_dir: Path | None = None,
    ) -> None: ...
    @property
    def backend(self) -> str: ...
    def compose_dependencies_line(self, suffix: str) -> Iterator[str]: ...
    def compose_optional_lines(self, suffix: str) -> Iterator[str]: ...
    def compose(self, suffix: str) -> str: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. py:data:: __all__
   :type: tuple[str]
   :value: ("ConfigSettings",)

   Module exports

.. py:data:: log
   :type: logging.Logger

   Module level logger

"""

import logging
import os
from pathlib import (
    Path,
    PurePath,
)

from .patch_pyproject_reading import ReadPyproject

__all__ = ("ConfigSettings",)
log = logging.getLogger("drain_swamp.monkey.config_settings")


class ConfigSettings:
    """There is no clear guidance on how build module should communicate
    ``config settings`` command line options to the build back end.

    This is the communication flow:

    .. code-block:: text

       - build
         --> setuptools
             --> build front end
             --> build backend (within a subprocess)
                 --> (plugin) manager
                     --> build plugin

    Requires a PEP, for all to agree, on how to pass these config settings
    to the build backend subprocess

    In the meantime, **this is the work around** for communication with
    the build backend

    - create a environment variable, DS_CONFIG_SETTINGS, set value to path to a temp file

    - temp file is in ``pyproject.toml`` format/style. ``project.name`` and
      ``project.version`` are required but not used. Contains a section, "config-settings".
      That section, contains key/value pairs. Type are both str.

      Which key value pairs to pass in depends on what is required by
      build plugins. Current key pairs:

      ``kind = "tag"``

      ``set-lock = "1"``

    :cvar SECTION_NAME: .toml file tool section name
    :vartype SECTION_NAME: str
    :cvar ENVIRON_KEY:

       Environment variable which stores the path to the temporary .toml
       file containing command line options intended for the build plugins

    :vartype ENVIRON_KEY: str
    :cvar FILE_NAME_DEFAULT: config settings temporary file name, not path
    :vartype FILE_NAME_DEFAULT: str
    :ivar file_name: Actual config settings temporary file name. Overrides default
    :vartype file_name: str | None
    """

    SECTION_NAME: str = "config-settings"
    ENVIRON_KEY: str = "DS_CONFIG_SETTINGS"
    FILE_NAME_DEFAULT: str = "setuptools-build.toml"

    __slots__ = ("_file_name",)

    def __init__(self, file_name=None):
        super().__init__()
        self.file_name = file_name

    @property
    def file_name(self):
        """Get config settings file name. This is a temporary file

        :returns: config settings file name
        :rtype: str
        """
        return self._file_name

    @file_name.setter
    def file_name(self, val):
        """Override default config file file name

        :param val: Expecting a str
        :type val: typing.Any
        """
        cls = type(self)
        is_none_or_empty = (
            val is None or not isinstance(val, str) or len(val.strip()) == 0
        )
        if is_none_or_empty:
            self._file_name = cls.FILE_NAME_DEFAULT
        else:
            self._file_name = val

    @classmethod
    def get_abs_path(cls):
        """From os.environ get temporary absolute file path

        :returns:
        :rtype: str | None
        """
        ret = os.environ.get(cls.ENVIRON_KEY, None)
        return ret

    @classmethod
    def set_abs_path(cls, val):
        """Set absolute temporary file path to config settings .toml file

        :param val: Expecting either str or Path. Should be absolute path
        :type val: typing.Any
        """
        if val is not None and issubclass(type(val), PurePath):
            toml_path = str(val)
        elif val is not None and isinstance(val, str):
            toml_path = val
        else:
            toml_path = None

        if toml_path is not None:
            os.environ[cls.ENVIRON_KEY] = toml_path
        else:  # pragma: no cover
            pass

    @classmethod
    def remove_abs_path(cls):
        """Remove environment variable holding the config settings temp file path"""
        del os.environ[cls.ENVIRON_KEY]

    def read(self):
        """Environment variable DS_CONFIG_SETTINGS contains an absolute
        path to a .toml file containing plugins' key/value pairs

        If :code:`python -m build` config setting cli options would be
        passed thru, by setuptools, this function would be unnecessary

        setuptools currently lacks this expected feature

        :returns:

           config settings dict. What normally would be supplied as
           :code:`python -m build` config setting cli options

        :rtype: dict[str, typing.Any]
        """
        cls = type(self)
        msg_warn = (
            "Expected env variable, DS_CONFIG_SETTINGS. Should contain "
            "path to .toml file. File contains project.name, project.version, "
            "tool.config-settings.kind and tool.config-settings.set-lock"
        )
        toml_path = cls.get_abs_path()
        if toml_path is not None:
            path_f = Path(toml_path)
            is_exists = path_f.exists() and path_f.is_file()
            if is_exists:
                """Attempt to read toml file. section tool.config-settings.
                Get all key/value pairs
                """
                try:
                    pyproj_data = ReadPyproject()(
                        path=path_f, tool_name=cls.SECTION_NAME
                    )
                except LookupError:
                    d_section = {}
                    log.warning(msg_warn)
                else:
                    # from setuptools_scm._integration.toml import TOML_RESULT
                    d_section = pyproj_data.section

                    # No config settings. Log warning
                    if len(d_section.keys()) == 0:
                        log.warning(msg_warn)
                    else:  # pragma: no cover
                        pass

            else:  # pragma: no cover
                d_section = {}
        else:  # pragma: no cover
            d_section = {}

        return d_section

    def write(
        self,
        path_dir,
        toml_contents,
    ):
        """The temp file will contain key/value pairs used by the build plugins.
        setuptools does not currently pass config_settings, requires a PEP
        for all to agree on how to pass these config settings from
        build --> setuptools --> build front end --> build backend in a subprocess.

        In the meantime, **this is the work around** for communication with
        the build backend

        :param path_dir:

           Will create a throw away file, so folder usually in /tmp or equivalent

        :type path_dir: pathlib.Path
        :param toml_contents:

           In ``pyproject.toml`` style, so project.name and project.version are
           required, but not used. Has section, ``tool.config-settings``. With
           key/value pairs. Both key and value are str.

           Which key value pairs to pass in depends on what is required by
           build plugins. Current key pairs:

           - ``kind = "tag"``

           - ``set-lock = "1"``

        :type toml_contents: str
        :param file_name:

           Default "setuptools-build.toml". File name of the temp .toml file

        :type file_name: str | None
        """
        cls = type(self)
        # overwrite ok. Not previously exists ok
        path_toml = path_dir.joinpath(self.file_name)
        path_toml.write_text(toml_contents)

        # environment variable contains path to a .toml file in a tmp folder
        # The temp file will contain key/value pairs used by the build plugins
        cls.set_abs_path(path_toml)

    @classmethod
    def get_section_dict(cls, path_dir, toml_contents, file_name=None):
        """Set the config settings and returning the section mapping

        :param path_dir:

           Will create a throw away file, so folder usually in /tmp or equivalent

        :type path_dir: pathlib.Path
        :param toml_contents:

           In ``pyproject.toml`` style, so project.name and project.version are
           required, but not used. Has section, ``tool.config-settings``. With
           key/value pairs. Both key and value are str.

           Which key value pairs to pass in depends on what is required by
           build plugins. Current key pairs:

           - ``kind = "tag"``

           - ``set-lock = "1"``

        :type toml_contents: str
        :param file_name:

           Default "setuptools-build.toml". File name of the temp .toml file

        :type file_name: str | None
        :returns:

           config settings dict. What normally would be supplied as
           :code:`python -m build` config setting cli options

        :rtype: dict[str, typing.Any]
        """
        cs = cls(file_name=file_name)
        cs.write(path_dir, toml_contents)
        d_section = cs.read()

        return d_section



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_package_metadata.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_package_metadata.py

Needs a config file to specify exact files to include / omit from report.
Will fail with exit code 1 even with 100% coverage

"""

from pathlib import Path

import pytest

from drain_swamp.package_metadata import (
    AUTHOR_NAME_FALLBACK,
    PackageMetadata,
    get_author_and_email,
    is_package_installed,
)


def test_play_with_cache():
    """If package is not installed, then author cache will contain only fallbacks"""
    app_name = "blarneys_blowup_doll"
    assert is_package_installed(app_name) is False
    author_name, author_email = get_author_and_email(app_name)
    assert author_name == AUTHOR_NAME_FALLBACK
    assert author_email is None


testdata_package_metadata_construct = (
    (
        "blarneys_blowup_doll",
        None,
        AUTHOR_NAME_FALLBACK,
        None,
    ),
    (
        "blarneys_blowup_doll",
        1.1234,
        AUTHOR_NAME_FALLBACK,
        None,
    ),
    (
        "blarneys_blowup_doll",
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        AUTHOR_NAME_FALLBACK,
        None,
    ),
    (
        "blarneys_blowup_doll",
        Path(__file__).parent.joinpath("_bad_files", "backend_only.pyproject_toml"),
        AUTHOR_NAME_FALLBACK,
        None,
    ),
    (
        "blarneys_blowup_doll",
        Path(__file__).parent.joinpath("_good_files", "no_project_name.pyproject_toml"),
        "Dave Faulkmore",
        "faulkmore@protonmail.com",
    ),
)

ids_package_metadata_construct = (
    "Unsupported type. Expected Path got None",
    "Unsupported type. Expected Path got float",
    "pyproject.toml lacks author name and email",
    "invalid pyproject.toml. Cannot parse",
    "pyproject.toml has author name and email",
)


@pytest.mark.parametrize(
    "app_name, path_config, expected_name, expected_email",
    testdata_package_metadata_construct,
    ids=ids_package_metadata_construct,
)
def test_package_metadata_construct(
    app_name,
    path_config,
    expected_name,
    expected_email,
    tmp_path,
    prep_pyproject_toml,
):
    # prepare
    path_pyproject_toml = prep_pyproject_toml(path_config, tmp_path)

    # app name cannot be None
    invalids = (
        None,
        "",
        "   ",
    )
    for invalid in invalids:
        with pytest.raises(ValueError):
            PackageMetadata(invalid, path=path_pyproject_toml)

    # nonexistent package. Just checked, it's available. Put the team on that right away

    pm = PackageMetadata(app_name, path=path_pyproject_toml)
    assert pm.full_name == expected_name
    assert pm.email == expected_email
    if pm.email is not None:
        assert isinstance(pm.d_pyproject_toml, dict)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

pep518 backends support ABC

Assumes package:

- uses ``pyproject.toml`` to store static packaging info. Especially not ``setup.cfg``.
  ``setup.py`` usage should be limited to C code and semantic versioning integration

- Not supporting py38-. Removes any justification for prefering
  ``setup.cfg`` or ``setup.py``

- python version isn't locked. e.g. ``requires-python = "~=3.9"``

.. py:data:: __all__
   :type: tuple[str]
   :value: ("BackendType",)

   Module's exports

.. py:data:: _logger
   :type: logging.Logger

   Module level logger

.. py:data:: is_module_debug
   :type: bool
   :value: False

   Flag to turn on module level logging

.. py:data:: entrypoint_name
   :type: str
   :value: "pipenv-unlock"

   Entrypoint name. For access to settings in ``pyproject.toml``

"""

from __future__ import annotations

import abc
import logging
import sys
from pathlib import (
    Path,
    PurePath,
)
from typing import (
    TYPE_CHECKING,
    ClassVar,
    cast,
)
from unittest.mock import MagicMock

from .check_type import (
    is_iterable_not_str,
    is_ok,
    is_relative_required,
)
from .constants import (
    SUFFIX_IN,
    SUFFIX_LOCKED,
    SUFFIX_SYMLINK,
    SUFFIX_UNLOCKED,
    g_app_name,
)
from .exceptions import (
    BackendNotSupportedError,
    MissingRequirementsFoldersFiles,
    PyProjectTOMLParseError,
    PyProjectTOMLReadError,
)
from .parser_in import TomlParser

if sys.version_info >= (3, 9):  # pragma: no cover
    from collections.abc import Sequence
else:  # pragma: no cover
    from typing import Sequence

__package__ = "drain_swamp"
__all__ = ("BackendType",)

_logger = logging.getLogger(f"{g_app_name}.backend_abc")

is_module_debug = False

# taken from pyproject.toml
entrypoint_name = "pipenv-unlock"  # noqa: F401


def try_dict_update(
    set_both,
    path_config,
    target_x,
    path_relative_x,
    is_bypass=False,
):
    """Helper to add absolute paths into pass by reference set

    ``pyproject.toml`` then cli. cli has higher priority

    :param set_both: pass by reference set. Changed in-place
    :type set_both: dict[str, pathlib.Path]
    :param path_config: base folder absolute path
    :type path_config: pathlib.Path
    :param target_x: target for relative path
    :type target_x: str
    :param path_relative_x: a relative path
    :type path_relative_x: pathlib.Path
    :param is_bypass: Default False. Should be a bool. Bypass exists and file check
    :type is_bypass: typing.Any | None
    """
    if is_bypass is None or not isinstance(is_bypass, bool):
        is_bypass = False
    else:  # pragma: no cover
        pass

    is_relative = is_relative_required(path_relative=path_relative_x)
    if is_relative:
        if not is_bypass:
            path_abs = path_config.joinpath(path_relative_x)
            is_file = path_abs.exists() and path_abs.is_file()
            if is_file:
                set_both.update({target_x: path_abs})
            else:  # pragma: no cover
                pass
        else:
            path_abs = path_config.joinpath(path_relative_x)
            set_both.update({target_x: path_abs})
    else:  # pragma: no cover
        pass


def get_optionals_cli(
    d_both,
    path_config,
    optionals,
):
    """Check sequence of relative paths, to a set, add absolute path(s)

    Helper to make testing possible
    :param d_both: target and absolute paths
    :type d_both: dict[str, pathlib.Path]
    :param path_config: absolute path folder
    :type path_config: pathlib.Path
    :param optionals:

       Preferably relative path contains Path or str. Check type and
       whether path is relative, exists, and is a file

    :type optionals: dict[str, typing.Any]
    """
    is_dict = optionals is not None and isinstance(optionals, dict)
    if is_dict:
        for target, opt in optionals.items():
            if opt is None:
                pass
            else:
                is_pathlike = isinstance(opt, str) or issubclass(type(opt), PurePath)
                if is_pathlike:
                    # supported types
                    path_opt = Path(opt)
                    try_dict_update(d_both, path_config, target, path_opt)
                else:  # pragma: no cover
                    # unsupported type
                    pass
        else:
            # empty sequence
            pass
    else:  # pragma: no cover
        pass


def get_optionals_pyproject_toml(
    d_both,
    d_pyproject_toml,
    path_config,
    is_bypass=False,
):
    """Get options from [tool.pipenv-unlock] optionals.
    dict key / value is target / relative path

    Change the rel path into an absolute path

    :param d_both: updated in-place. key / value --> target / absolute path
    :type d_both: dict[str, pathlib.Path]
    :param d_pyproject_toml: pyproject.toml dict
    :type d_pyproject_toml: dict[str, typing.Any]
    :param path_config: absolute folder path
    :type path_config: pathlib.Path
    :param is_bypass: Default False. True checks if file exists
    :type is_bypass: bool | None
    """
    if is_bypass is None or not isinstance(is_bypass, bool):
        is_bypass = False

    d_tool = d_pyproject_toml.get("tool", {}).get(entrypoint_name, {})
    for key, mixed_blob in d_tool.items():
        is_optionals = (
            key == "optionals"
            and mixed_blob is not None
            and isinstance(mixed_blob, Sequence)
        )
        if is_module_debug:  # pragma: no cover
            _logger.info(f"is_optionals: {is_optionals}")
        else:  # pragma: no cover
            pass

        if is_optionals:
            for d_optional_dependency in mixed_blob:
                if "target" in d_optional_dependency.keys():
                    target = d_optional_dependency["target"]
                else:  # pragma: no cover
                    target = None

                if "relative_path" in d_optional_dependency.keys():
                    opt = d_optional_dependency["relative_path"]
                else:  # pragma: no cover
                    opt = None

                is_missing_expected_keys = target is None or opt is None
                if is_missing_expected_keys:  # pragma: no cover
                    """a dict, but no recognized keys. Expected target
                    and relative_path"""
                    continue
                else:  # pragma: no cover
                    pass

                is_pathlike = isinstance(opt, str) or issubclass(type(opt), PurePath)
                if is_module_debug:  # pragma: no cover
                    msg_info = f"is_pathlike: {is_pathlike}"
                    _logger.info(msg_info)
                else:  # pragma: no cover
                    pass

                if is_pathlike:
                    path_opt = Path(opt)
                    if is_module_debug:  # pragma: no cover
                        msg_info = f"target / rel path: {target} / {path_opt}"
                        _logger.info(msg_info)
                    else:  # pragma: no cover
                        pass

                    try_dict_update(
                        d_both,
                        path_config,
                        target,
                        path_opt,
                        is_bypass=is_bypass,
                    )
                else:  # pragma: no cover
                    # unsupported type
                    pass
            else:
                # empty sequence
                pass
        else:  # pragma: no cover
            pass


def get_required_pyproject_toml(
    d_pyproject_toml,
    path_config,
    is_bypass=False,
):
    """From pyproject.toml [tool.pipenv-unlock], get ``required`` value.
    Which is a dict

    is_bypass controls whether it must already exist or not. Useful for testing

    :param d_pyproject_toml: pyproject.toml dict
    :type d_pyproject_toml: dict[str, typing.Any]
    :param path_config: absolute folder path
    :type path_config: pathlib.Path
    :param is_bypass: Default False. True checks if file exists
    :type is_bypass: bool | None
    :returns: tuple containing required target and abs path
    :rtype: tuple[str, pathlib.Path] | None
    """
    if is_bypass is None or not isinstance(is_bypass, bool):
        is_bypass = False

    ret = None
    d_tool = d_pyproject_toml.get("tool", {}).get(entrypoint_name, {})
    for key, mixed_blob in d_tool.items():
        is_required = (
            key == "required"
            and mixed_blob is not None
            and isinstance(mixed_blob, dict)
        )
        if is_module_debug:  # pragma: no cover
            _logger.info(f"is_required: {is_required}")
        else:  # pragma: no cover
            pass

        if is_required:
            is_target = "target" in mixed_blob.keys()
            is_relative_path = "relative_path" in mixed_blob.keys()
            if is_module_debug:  # pragma: no cover
                _logger.info(f"is_target (pyproject.toml): {is_target}")
                _logger.info(f"is_relative_path (pyproject.toml): {is_relative_path}")
            else:  # pragma: no cover
                pass
            if is_target and is_relative_path:
                target_b = cast(str, mixed_blob["target"])
                path_rel_b = cast(str, mixed_blob["relative_path"])
                # assumes default extensions
                is_relative = is_relative_required(path_relative=Path(path_rel_b))
                if is_module_debug:  # pragma: no cover
                    _logger.info(f"is_relative (pyproject.toml): {is_relative}")
                else:  # pragma: no cover
                    pass
                if is_relative:
                    path_abs = path_config.joinpath(path_rel_b)
                    if is_bypass:
                        ret = (target_b, path_abs)
                    else:
                        is_file = path_abs.exists() and path_abs.is_file()
                        if is_module_debug:  # pragma: no cover
                            _logger.info(f"is_file (pyproject.toml): {is_file}")
                        else:  # pragma: no cover
                            pass
                        if is_file:
                            ret = (target_b, path_abs)
                        else:
                            # not a file
                            ret = None
                else:  # pragma: no cover
                    # unexpected absolute path
                    pass
            else:  # pragma: no cover
                """dict does not contain both target and relative_path
                keys and respective values"""
                pass
        else:  # pragma: no cover
            # in pyproject.toml [tool.pipenv-unlock], no required key nor dict value
            pass

    return ret


def get_required_cli(
    path_config,
    required=None,
    is_bypass=False,
):
    """Process required field provided by cli

    :param path_config: absolute folder path
    :type path_config: pathlib.Path
    :param required: required package dependencies provided by cli
    :type required: tuple[str, typing.Any] | None
    :param is_bypass: Default False. True checks if file exists
    :type is_bypass: bool | None
    :returns:

       None if no suitable pair otherwise target and abs path

    :rtype: tuple[str, pathlib.Path] | None
    """
    if is_bypass is None or not isinstance(is_bypass, bool):
        is_bypass = False

    ret = None
    if required is not None and is_iterable_not_str(required):
        target_a, path_a = required
        if is_ok(path_a):
            # Check path ok
            is_relative = is_relative_required(path_relative=Path(path_a))
            if is_relative:
                path_abs = path_config.joinpath(path_a)
                if is_bypass:
                    ret = (target_a, path_abs)
                else:
                    is_file = path_abs.exists() and path_abs.is_file()
                    if is_file:
                        ret = (target_a, path_abs)
                    else:  # pragma: no cover
                        # not a file
                        pass
            else:  # pragma: no cover
                # unexpected absolute path
                pass
        elif issubclass(type(path_a), PurePath):
            is_relative = is_relative_required(path_relative=path_a)
            if is_relative:
                path_abs = path_config / path_a
                if is_bypass:
                    ret = (target_a, path_abs)
                else:
                    is_file = path_abs.exists() and path_abs.is_file()
                    if is_file:
                        ret = (target_a, path_abs)
                    else:  # pragma: no cover
                        # not a file
                        pass
            else:  # pragma: no cover
                # unexpected absolute path
                pass
        else:  # pragma: no cover
            # cli required path is unsupported type; do nothing
            pass
    else:
        pass

    return ret


def folders_implied_init(
    parent_dir,
    optionals,
    required=None,
):
    """Determine implied folders.

    Set of relative paths to folders containing ".in" files

    Determined from these properties:

    - type[BackendType].required
    - type[BackendType].optionals

    :param parent_dir: absolute path to parent folder
    :type parent_dir: pathlib.Path
    :param optionals: optional dependencies. target / absolute path
    :type optionals: dict[str, pathlib.Path]
    :param required: Default None. required dependencies. target / absolute path
    :type required: tuple[str, Path] | None
    :returns: set of **relative paths** to folders containing ``.in`` files
    :rtype set[pathlib.Path]
    """
    set_folders_implied = set()
    if required is not None:
        target_a, abs_path_a = required
        path_relative_a = abs_path_a.relative_to(parent_dir)
        path_rel_dir_a = path_relative_a.parent
        if path_rel_dir_a != Path("."):
            set_folders_implied.add(path_rel_dir_a)
        else:  # pragma: no cover
            pass
    else:  # pragma: no cover
        pass

    for target_b, abs_path_b in optionals.items():
        path_relative_b = abs_path_b.relative_to(parent_dir)
        path_rel_dir_b = path_relative_b.parent
        if path_rel_dir_b != Path("."):
            set_folders_implied.add(path_rel_dir_b)
        else:  # pragma: no cover
            pass

    return set_folders_implied


def folders_additional_init(
    parent_dir,
    folders_implied,
    additional_folders=(),
):
    """Additional folders that cannot be determined solely from the dependences or
    optional-dependences. These additional requirements are typically for:

    - tox

    - mypy

    - CI/CD

    In ``pyproject.toml``,

    .. code-block:: text

       [tool.setuptools.dynamic]
       # @@@ editable little_shop_of_horrors_shrine_candles
       dependencies = { file = ["requirements/prod.lock"] }
       optional-dependencies.pip = { file = ["requirements/pip.lock"] }
       optional-dependencies.pip_tools = { file = ["requirements/pip-tools.lock"] }
       optional-dependencies.dev = { file = ["requirements/dev.lock"] }
       optional-dependencies.manage = { file = ["requirements/manage.lock"] }
       optional-dependencies.docs = { file = ["docs/requirements.lock"] }
       # @@@ end

       [tool.pipenv-unlock]
       folders = [
           "docs",
           "requirements",
           "ci",
       ]

       required = { target = "prod", relative_path = "requirements/prod.in" }


       optionals = [
           { target = "pip", relative_path = "requirements/pip.in" },
           { target = "pip_tools", relative_path = "requirements/pip-tools.in" },
           { target = "dev", relative_path = "requirements/dev.in" },
           { target = "manage", relative_path = "requirements/manage.in" },
           { target = "docs", relative_path = "docs/requirements.in" },
       ]

    The implied folders are determined from:

    - tool.pipenv-unlock.required
    - tool.pipenv-unlock.optionals

    In this case, the implied folders are: :code:`{"requirements", "docs"}`

    tool.pipenv-unlock.folders contains the additional folder: :code:`{"ci"}`

    :param parent_dir: absolute path to parent folder
    :type parent_dir: pathlib.Path
    :param folders_implied: Relative path of folders that contain ``.in`` files
    :type folders_implied: set[pathlib.Path]
    :param additional_folders:

       Relative path of additional folders that contain ``.in`` files

    :type additional_folders: tuple[pathlib.Path]
    :returns: set of **relative paths** to additional folders that contain ``.in`` files
    :rtype set[pathlib.Path]
    """
    set_folders_additional = set()
    for path_mixed in additional_folders:
        # click may resolve Path --> str
        if isinstance(path_mixed, str):
            path_mixed = Path(path_mixed)
        else:  # pragma: no cover
            pass

        if path_mixed.is_absolute():
            if (
                path_mixed.exists()
                and path_mixed.is_dir()
                and path_mixed.is_relative_to(parent_dir)
            ):
                path_relative_c = path_mixed.relative_to(parent_dir)
                # folders_implied contains relpath
                is_additional = path_relative_c not in folders_implied
                if path_relative_c != Path(".") and is_additional:
                    set_folders_additional.add(path_relative_c)
                else:  # pragma: no cover
                    # same as parent dir
                    pass
            else:  # pragma: no cover
                # doesn't exist, or not a folder, or not relative to parent folder
                pass
        else:
            abs_path_c = parent_dir / path_mixed
            # folders_implied contains relpath
            is_additional = path_mixed not in folders_implied
            if abs_path_c.exists() and abs_path_c.is_dir() and is_additional:
                set_folders_additional.add(path_mixed)

    return set_folders_additional


def ensure_folder(val):
    """
    :param val:

       Preferrable a Path. Can be either a file or a folder. We want the folder

    :type val: typing.Any
    :returns: A folder Path
    :rtype: pathlib.Path
    :raises:

       - :py:exc:`NotADirectoryError` -- BackendType.path_config
         is neither a file nor a folder

    """
    msg_exc = "Expecting a file or a folder path. Received something else"
    path_v = None
    if val is None:
        path_val = None
    else:
        if isinstance(val, MagicMock):
            # For testing only. Expect the return_value to be a Path
            path_v = val.return_value
            if issubclass(type(val.return_value), PurePath):
                path_val = path_v
            else:  # pragma: no cover
                pass
        elif issubclass(type(val), PurePath):
            path_val = val
        else:
            path_val = None

    # Two birds one stone
    if path_val is None:
        # None or a MagicMock.return_value which is not a Path
        msg_exc = f"Unsupported type got {type(val)}"
        raise TypeError(msg_exc)
    else:  # pragma: no cover
        pass

    if path_val.is_file() and not path_val.is_symlink():
        ret = path_val.parent
    elif path_val.is_dir() and not path_val.is_symlink():
        ret = path_val
    else:
        """symlink, is junction (Windows only), socket, fifo,
        block device, or char device"""
        raise NotADirectoryError(msg_exc)

    return ret


class BackendType(abc.ABC):
    """ABC of packaging backend support.

    get_registered and __subclasshook__ makes this ABC aware of all subclasses

    All that is required is to import this ABC and all subclasses

    To create a subclass backend instance, call
    :py:meth:`drain_swamp.backend_abc.BackendType.load_factory`.

    The backend is known immediately after reading in ``pyproject.toml``
    and the appropriate subclass is chosen and instanciated
    """

    _path_required: tuple[str, Path] | None
    _paths_optional: dict[str, Path]
    _path_config: Path
    _parent_dir: Path
    _folders_implied: set[Path]
    _folders_additional: set[Path]
    BACKEND_NAME: ClassVar[str]
    PYPROJECT_TOML_SECTION_NAME: ClassVar[str]

    @staticmethod
    def read(
        path_config,
    ):
        """Read the current contents of ``pyproject.toml`` file

        :param path_config: ``pyproject.toml`` folder path
        :type path_config: pathlib.Path
        :returns: ``pyproject.toml`` dict and resolved path to file
        :rtype: tuple[dict[str, typing.Any], pathlib.Path]
        :raises:

           - :py:exc:`drain_swamp.exceptions.PyProjectTOMLParseError` --
             either not found or cannot be parsed

           - :py:exc:`drain_swamp.exceptions.PyProjectTOMLReadError` --
             Either not a file or lacks read permission

        """
        # Expects a Path. get_pyproject_toml call won't create a TypeError
        if is_module_debug:  # pragma: no cover
            _logger.info(f"BackendType.read path_config: {path_config!r}")
        else:  # pragma: no cover
            pass

        is_type_ng = path_config is None or not (
            isinstance(path_config, str) or issubclass(type(path_config), PurePath)
        )

        # Avoids TypeError during resolve_pyproject_toml call
        if is_type_ng:
            msg_exc = "pyproject.toml is either not a file or lacks r/w permission"
            raise PyProjectTOMLReadError(msg_exc)
        else:  # pragma: no cover
            pass

        try:
            # raise TypeError, FileNotFoundError, or PyProjectTOMLParseError
            tp = TomlParser(
                path_config,
                raise_exceptions=True,
            )
            d_pyproject_toml = tp.d_pyproject_toml
            path_f = tp.path_file
        except (PyProjectTOMLParseError, PyProjectTOMLReadError):
            raise

        return d_pyproject_toml, path_f

    @staticmethod
    def load_factory(
        path_config,
        required=None,
        optionals={},
        parent_dir=None,
        additional_folders=(),
    ):
        """Choose factory from registered subclasses

        :param path_config: ``pyproject.toml`` folder path
        :type path_config: pathlib.Path
        :param required: relative path to required dependency .in file
        :type required: tuple[str, pathlib.Path] | None
        :param optionals:

           Default empty tuple. Relative path to optional dependencies
           .in files. There may be some which are not optional dependencies
           like ``requirements/tox.in`` and ``requirements/kit.in``. Which
           are used by CI/CD or tox

        :type optionals: dict[str, pathlib.Path]
        :param parent_dir:

           folder path If provided and acceptable overrides attr, path_config

        :type parent_dir: pathlib.Path | None
        :param additional_folders:

           Default empty Sequence. Folders to search for .in files beyond
           the folders implied by required and optionals relative_path values

        :type additional_folders: tuple[pathlib.Path, ...]
        :returns: Subclass instance
        :rtype: typing.Self
        :raises:

           - :py:exc:`drain_swamp.exceptions.PyProjectTOMLReadError` --
             Either not a file or lacks read permission

           - :py:exc:`drain_swamp.exceptions.PyProjectTOMLParseError` --
             Could not parse pyproject.toml for various reasons

           - :py:exc:`drain_swamp.exceptions.BackendNotSupportedError` --
             No support yet for python package backend

        """

        # During testing, path_config can be cwd, while parent_dir is tmp_path
        if parent_dir is not None and issubclass(type(parent_dir), PurePath):
            path_override = parent_dir
        else:
            path_override = path_config

        if is_module_debug:  # pragma: no cover
            msg_info = (
                "drain_swamp.backend_abc.BackendType:load_factory "
                f"path_override {path_override!r}"
            )
            _logger.info(msg_info)
        else:  # pragma: no cover
            pass

        # May raise PyProjectTOMLParseError or PyProjectTOMLReadError
        d_pyproject_toml, path_f = BackendType.read(path_override)

        str_backend = BackendType.determine_backend(d_pyproject_toml)
        lst_registered = list(BackendType.get_registered())

        ret = None
        for kls in lst_registered:
            if kls.BACKEND_NAME == str_backend:
                ret = kls(  # type: ignore[call-arg]
                    d_pyproject_toml,
                    path_f,
                    required=required,
                    optionals=optionals,
                    parent_dir=parent_dir,
                    additional_folders=additional_folders,
                )
            else:  # pragma: no cover
                # continue
                pass

        if ret is None:
            backends_name = [kls.BACKEND_NAME for kls in lst_registered]
            msg_exc = (
                f"No support yet for python package backend {str_backend}. "
                f"Registered backends: {backends_name}. The pyproject.toml "
                f"{path_override}, specify backend in build-system.build-backend "
                "or override tool.pipenv-unlock.wraps-build-backend"
            )
            raise BackendNotSupportedError(msg_exc)
        else:  # pragma: no cover
            pass

        return ret

    def load(
        self,
        d_pyproject_toml,
        required=None,
        optionals={},
        parent_dir=None,
        additional_folders=(),
    ):
        """Called by subclass to process both cli and ``pyproject.toml`` values

        :param d_pyproject_toml: pyproject.toml dict
        :type d_pyproject_toml: dict[str, typing.Any],
        :param required:

           Default None. Required dependency relative path to .in file. If exists and
           is a file will override pyproject.toml respective value

        :type required: tuple[str, pathlib.Path] | None
        :param optionals:

           Default empty dict. Relative path to optional dependencies
           ``.in`` files. There may be some which are not optional dependencies
           like ``requirements/tox.in`` and ``requirements/kit.in``. Which
           are used by CI/CD or tox

        :type optionals: dict[str, pathlib.Path]
        :param parent_dir:

           folder path If provided and acceptable overrides attr, path_config

        :type parent_dir: pathlib.Path | None
        :param additional_folders:

           Default empty Sequence. Folders to search for .in files beyond
           the folders implied by required and optionals relative_path values

        :type additional_folders: tuple[pathlib.Path, ...]
        """

        # May raise NotADirectoryError, if path_config is neither a file or a folder
        # AND parent_dir not provided
        self.parent_dir = parent_dir

        # path_required Path
        t_c = BackendType.get_required(
            d_pyproject_toml,
            self._parent_dir,
            required=required,
        )
        if t_c is None:
            # Package has no requires
            self._path_required = None
        else:
            self._path_required = t_c

        #    file extension assumed to be .in``
        #    There being no optionals is strange, *but* not an error
        self._paths_optional = BackendType.get_optionals(
            d_pyproject_toml,
            self.parent_dir,
            optionals,
        )

        # _folders_implied
        self._folders_implied = folders_implied_init(
            self.parent_dir,
            self.optionals,
            required=required,
        )

        # _folders_additional
        self._folders_additional = folders_additional_init(
            self.parent_dir,
            self.folders_implied,
            additional_folders=additional_folders,
        )
        #    Annoying to make a unittest just for this property
        assert self.folders_additional == self._folders_additional

    @staticmethod
    def determine_backend(d_pyproject_toml):
        """If a thin-wrapped custom build backend, must specify build
        backend

        .. code-block:: text

           [tool.pipenv-unlock]
           wraps-build-backend = "setuptools.build_meta"

        Example custom build backend

        .. code-block:: text

           [build-system]
           requires = ["setuptools>=70.0.0", "wheel", "build", "setuptools_scm>=8"]
           build-backend = "_req_links.backend"
           backend-path = [
               ".",
               "src",
           ]

        :param d_pyproject_toml: ``pyproject.toml`` read in as a dict
        :type d_pyproject_toml: dict[str, typing.Any]
        :returns:

           backend package name. Up to first period. e.g.
           setuptools.build_meta --> "setuptools"

        :rtype: str

        .. seealso::

           https://setuptools.pypa.io/en/latest/build_meta.html#dynamic-build-dependencies-and-other-build-meta-tweaks

        """
        str_build_system = (
            d_pyproject_toml.get("tool", {})
            .get("pipenv-unlock", {})
            .get("wraps-build-backend", None)
        )

        if str_build_system is None:
            # build-backend not thin-wrapped
            d_build_system = d_pyproject_toml.get("build-system", {})
            str_build_system = d_build_system["build-backend"]
        else:  # pragma: no cover
            pass

        str_build = str_build_system.split(".")[0]
        ret = cast(str, str_build)

        return ret

    @staticmethod
    def get_required(
        d_pyproject_toml,
        path_config,
        required=None,
    ):
        """From pyproject.toml, retrieve which required dependencies

        cli priority over pyproject.toml

        :param d_pyproject_toml: ``pyproject.toml`` read in as a dict
        :type d_pyproject_toml: dict[str, typing.Any]
        :param path_config: absolute path to a folder
        :type path_config: pathlib.Path
        :param required:

           Default None. ``.in`` file path. Contents are the required dependencies

        :type required: tuple[str, typing.Any] | None
        :returns:

           None if no suitable pair otherwise target and abs path

        :rtype: tuple[str, pathlib.Path] | None
        """
        ret = get_required_pyproject_toml(
            d_pyproject_toml,
            path_config,
            is_bypass=False,
        )
        if is_module_debug:  # pragma: no cover
            _logger.info(f"required (pyproject.toml): {ret}")
        else:  # pragma: no cover
            pass

        # cli priority over pyproject.toml
        ret_cli = get_required_cli(
            path_config,
            required,
        )
        if is_module_debug:  # pragma: no cover
            _logger.info(f"required (cli): {ret_cli}")
        else:  # pragma: no cover
            pass

        if ret_cli is not None:
            ret = ret_cli
        else:  # pragma: no cover
            # pyproject.toml supplied required
            pass

        if is_module_debug:  # pragma: no cover
            _logger.info(f"required (final): {ret}")
        else:  # pragma: no cover
            pass

        return ret

    @staticmethod
    def get_optionals(
        d_pyproject_toml,
        path_config,
        optionals,
    ):
        """Combined cli and pyproject.toml optionals. Does not check relative path(s)

        cli supplements ``pyproject.toml``, not overrides completely

        :param d_pyproject_toml: ``pyproject.toml`` read in as a dict
        :type d_pyproject_toml: dict[str, typing.Any]
        :param path_config: absolute path to a folder
        :type path_config: pathlib.Path
        :param optionals: hopefully sequence of relative paths
        :type optionals: dict[str, typing.Any]
        :returns: tuple of paths absolute paths
        :rtype: dict[str, pathlib.Path]
        """
        if TYPE_CHECKING:
            dict_pairs: dict[str, Path]

        dict_pairs = {}

        # pyproject.toml tool.[package name].optionals
        get_optionals_pyproject_toml(
            dict_pairs,
            d_pyproject_toml,
            path_config,
        )

        # cli priority over pyproject.toml
        get_optionals_cli(
            dict_pairs,
            path_config,
            optionals,
        )

        return dict_pairs

    @property
    @abc.abstractmethod
    def backend(self):
        """Get the backend. This is extracted from pyproject.toml

        :returns: pep518 backend
        :rtype: str
        """
        ...

    @property
    def path_config(self):
        """Path to config file

        :returns: Path to config file
        :rtype: pathlib.Path
        """
        return self._path_config

    @property
    def parent_dir(self):
        """override folder Path. Likely a temp folder

        :returns: Path override folder
        :rtype: pathlib.Path
        """
        return self._parent_dir

    @parent_dir.setter
    def parent_dir(
        self,
        parent_dir=None,
    ):
        """path_config is supposed to be the folder containing
        ``pyproject.toml``, but it's unclear. Could be either a file or
        a folder.

        Sometimes prefer to override the folder location and this
        what parent_dir is intended for

        :param parent_dir:

           Default None. A folder, often a temp folder, that overrides
           :py:attr:`BackendType.path_config <drain_swamp.backend_abc.BackendType.path_config>`

        :type parent_dir: pathlib.Path
        :raises:

           - :py:exc:`NotADirectoryError` -- BackendType.path_config
             is neither a file nor a folder

        """

        # Avoid calling ensure_folder if parent_dir is acceptable
        if parent_dir is None:
            # may raise NotADirectoryError
            path_config = self.path_config
            path_dir = ensure_folder(path_config)
            if is_module_debug:  # pragma: no cover
                msg_info = (
                    "path_dir (None) --> self.path_config ("
                    f"{type(self.path_config)}): {path_dir}"
                )
                _logger.info(msg_info)
            else:  # pragma: no cover
                pass
        else:
            if not issubclass(type(parent_dir), PurePath):
                # may raise NotADirectoryError
                path_config = self.path_config
                path_dir = ensure_folder(path_config)
                if is_module_debug:  # pragma: no cover
                    msg_info = (
                        f"BackendType.parent_dir setter. path_dir: {path_dir!r} "
                        f"path config: ({self.path_config!r})"
                    )
                    _logger.info(msg_info)
                else:  # pragma: no cover
                    pass
            else:
                is_abs_dir = parent_dir.is_absolute() and parent_dir.is_dir()
                if is_abs_dir:
                    # override acceptable
                    path_dir = parent_dir
                    if is_module_debug:  # pragma: no cover
                        _logger.info(f"path_dir --> parent_dir: {path_dir}")
                    else:  # pragma: no cover
                        pass
                else:
                    # fallback override rejected may raise NotADirectoryError
                    path_config = self.path_config
                    path_dir = ensure_folder(path_config)

                    if is_module_debug:  # pragma: no cover
                        msg_info = (
                            "path_dir (fallback)--> self.path_config("
                            f"{type(self.path_config)}): {path_dir}"
                        )
                        _logger.info(msg_info)
                    else:  # pragma: no cover
                        pass

        self._parent_dir = path_dir

    @property
    def required(self):
        """Required dependency

        :returns: dependency name and absolute path
        :rtype: tuple[str, pathlib.Path] | None
        """
        return self._path_required

    @property
    def optionals(self):
        """Absolute paths to optional dependencies ``*.in`` files. Some
        may later not become an extra

        :returns: optional dependency name and absolute path to ``*.in`` files
        :rtype: dict[str, pathlib.Path]
        """
        return self._paths_optional

    @property
    def folders_implied(self):
        """Relative Path to folders containing requirements ``.in`` files

        Derived from paths in both required and optional dependencies

        :returns: Unique set of relative Path to folders
        :rtype: set[pathlib.Path]
        """
        return self._folders_implied

    @property
    def folders_additional(self):
        """Relative Path to folders which cannot be derived from
        required and optional dependencies alone.

        :returns: Unique set of relative Path to folders
        :rtype: set[pathlib.Path]
        """
        return self._folders_additional

    @classmethod
    def __subclasshook__(cls, C):
        """A class wanting to be
        :py:class:`~logging_strict.logging_yaml_abc.LoggingYamlType`,
        minimally requires:

        Properties:

        - file_stem

        - file_name

        - package

        - dest_folder

        Methods:

        - extract

        - as_str -- get for free

        - setup -- get for free

        Then register itself
        :code:`LoggingYamlType.register(AnotherDatumClass)` or subclass
        :py:class:`~logging_strict.logging_yaml_abc.LoggingYamlType`

        :param C:

           Class to test whether implements this interface or is a subclass

        :type C: typing.Any
        :returns:

           ``True`` implements
           :py:class:`~logging_strict.logging_yaml_abc.LoggingYamlType`
           interface or is a subclass. ``False`` not a
           :py:class:`~logging_strict.logging_yaml_abc.LoggingYamlType`

        :rtype: bool
        """
        if cls is BackendType:
            methods = ("compose",)

            expected_count = len(methods)
            for B in C.__mro__:
                lst = [True for meth in methods if meth in B.__dict__]
                match_count = len(lst)
                is_same = match_count == expected_count
                if is_same:
                    return True
                else:  # pragma: no cover
                    pass
        else:  # pragma: no cover
            pass
        return NotImplemented  # pragma: no cover Tried long enough with issubclass

    @classmethod
    def get_registered(cls):
        """The :py:mod:`abc` registry for abc
        :py:class:`BackendType`, contains registered classes.

        :returns: Generator of registered classes to abc, BackendType
        :rtype: collections.abc.Iterator[type[typing.Self]]

        .. note:: test howto

           Would have to create a mock with all the abstract properties
           and methods implemented. This is **alot** of work. Testing
           is skipped

        .. note:: :py:func:`abc._get_dump`

           Does exist in :py:mod:`abc`. The module level function is
           protected and not intended to be user facing.

           This is the only way to know which classes are registered to
           support this abc. Would prefer not to hard-code the
           registered classes

        """
        for weakref_cls in abc._get_dump(BackendType)[1]:  # type: ignore
            ref = weakref_cls
            klass = ref()
            if klass is not None:
                yield klass
            else:  # pragma: no cover
                pass

        yield from ()

    @staticmethod
    def fix_suffix(suffix):
        """Prepend period if suffix lacks it

        :param suffix:

           one file suffix, not suffixes, which might not be prepended by a period

        :type suffix: str
        :returns: Suffix with preceding period character
        :rtype: str
        """
        if not suffix.startswith("."):
            ret = f".{suffix}"
        else:
            ret = suffix

        return ret

    def in_files(self):
        """In both implied and additional folders, create generator of all ``.in`` files

        :returns: All the ``.in`` files in any of the implied and additional folders
        :rtype: collections.abc.Generator[pathlib.Path, None, None]
        """
        # combine two sets (of absolute paths)
        mod_path = "backend_abc.BackendType.in_files"
        set_folders = self.folders_implied | self.folders_additional
        if is_module_debug:  # pragma: no cover
            _logger.info(f"{mod_path} implied folders: {self.folders_implied}")
            _logger.info(f"{mod_path} additional folders: {self.folders_additional}")
            _logger.info(f"{mod_path} set_folders: {set_folders}")
        else:  # pragma: no cover
            pass

        for rel_path in set_folders:
            if is_module_debug:  # pragma: no cover
                _logger.info(f"{mod_path} rel_path: {rel_path}")
            else:  # pragma: no cover
                pass
            path_dir = self.parent_dir
            abs_path = path_dir.joinpath(rel_path)
            pattern = f"**/*{SUFFIX_IN}"
            if is_module_debug:  # pragma: no cover
                _logger.info(f"{mod_path} abs_path: {abs_path}")
                _logger.info(f"{mod_path} pattern {pattern}")
                _logger.info(f"""{mod_path} files: {list(abs_path.glob(pattern))}""")
            else:  # pragma: no cover
                pass
            yield from abs_path.glob(f"**/*{SUFFIX_IN}")
        yield from ()

    @staticmethod
    def is_locked(path_config):
        """pyproject.toml exists and the backend is supported. Check
        package dependency lock state.

        :param path_config: absolute path to ``pyproject.toml``
        :type path_config: pathlib.Path
        :returns:

           True if has mostly dependencies and optional dependencies
           have mostly ``.lock`` suffixes otherwise False

           None indicates snippet with that id does not exist

        :rtype: bool
        :raises:

           - :py:exc:`AssertionError` -- in pyproject.toml [tool.setuptools.dynamic]
              section, expect this section and at least key, dependencies

           - :py:exc:`drain_swamp.exceptions.PyProjectTOMLParseError` --
             either not found or cannot be parsed

           - :py:exc:`drain_swamp.exceptions.PyProjectTOMLReadError` --
             Either not a file or lacks read permission

        .. todo:: requirements.txt usage

           pip-compile usage encourages: .in and .txt
           .txt is equivalent to .lock, but .lock meaning is explicit and clear

           drain_swamp broke .txt --> .unlock and .lock

           What if accidently or still using .txt?

        """
        # tomli parser ignores snippet (comments).
        # This is for the best; parsing quotes in regex is too hard
        # Possible to get a snippet from an invalid ``pyproject.toml`` file
        # May raise PyProjectTOMLParseError or PyProjectTOMLReadError
        if is_module_debug:  # pragma: no cover
            msg_info = f"BackendType.is_locked reading path config: {path_config!r}"
            _logger.info(msg_info)
        else:  # pragma: no cover
            pass

        try:
            d_pyproject_toml, path_f = BackendType.read(path_config)
        except (PyProjectTOMLParseError, PyProjectTOMLReadError):
            raise

        if is_module_debug:  # pragma: no cover
            _logger.info(f"BackendType.is_locked d_pyproject_toml: {d_pyproject_toml}")
        else:  # pragma: no cover
            pass

        locks = []
        unlocks = []

        def sorting_hat(relpath: Path) -> None:
            nonlocal locks
            nonlocal unlocks
            nonlocal path_f

            file_name = relpath.name
            if file_name.endswith(SUFFIX_LOCKED):
                locks.append(relpath)
            elif file_name.endswith(SUFFIX_UNLOCKED):
                unlocks.append(relpath)
            elif file_name.endswith(SUFFIX_SYMLINK):
                # resolve symlink gets the file towhich it points
                abspath_package = path_f.parent
                abspath_dependency_file = abspath_package / relpath
                path_resolved = abspath_dependency_file.resolve()

                dependency_file_name = path_resolved.name
                if dependency_file_name.endswith(SUFFIX_LOCKED):
                    locks.append(path_resolved)
                elif dependency_file_name.endswith(SUFFIX_UNLOCKED):
                    unlocks.append(path_resolved)
                else:  # pragma: no cover
                    pass
            else:  # pragma: no cover
                pass

        def choose_winner() -> bool:
            """Choose winner

            Assumes suffixes are either: .unlock or .lock

            A file suffix of ``.txt`` is not supported

            :returns: True is locked otherwise False
            :rtype: bool
            """
            nonlocal locks
            nonlocal unlocks

            locks_count = len(locks)
            unlocks_count = len(unlocks)

            is_both_zero = locks_count == 0 and unlocks_count == 0
            if is_both_zero:  # pragma: no cover
                # both 0 ! Yikes!! No locks means is unlocked. Even if no unlocks
                ret = False
            else:
                is_locks_win = locks_count != 0
                if is_locks_win:
                    # locked
                    ret = True
                else:
                    # Not locked, so False
                    ret = False

            return ret

        section_ = (
            d_pyproject_toml.get("tool", {}).get("setuptools", {}).get("dynamic", {})
        )
        is_no_tool_setuptools_dynamic_section = len(section_.keys()) == 0

        required_count = 0
        for target, d_ in section_.items():
            if target == "dependencies":
                # required dependencies
                files = d_.get("file", [])
                for f_relpath in files:
                    required_count += 1
                    sorting_hat(Path(f_relpath))
            elif target == "optional-dependencies":
                # optional dependencies
                for target_opt, d_opt in section_.items():
                    files = d_opt.get("file", [])
                    for f_relpath in files:
                        sorting_hat(Path(f_relpath))
            else:  # pragma: no cover
                pass

        if is_no_tool_setuptools_dynamic_section or required_count != 1:
            """Minimally a package has one dependencies ``.in`` with
            cooresponding ``.unlock`` and ``.lock``

            .. code-block:: text

               [tool.setuptools.dynamic]
               # @@@ editable little_shop_of_horrors_shrine_candles
               dependencies = { file = ["requirements/prod.unlock"] }
               # @@@ end

            """
            msg_exc = (
                "in pyproject.toml [tool.setuptools.dynamic] section, "
                "expect this section and at least key, dependencies"
            )
            raise AssertionError(msg_exc)

        if is_module_debug:  # pragma: no cover
            _logger.info(f"BackendType.is_locked locks: {locks}")
            _logger.info(f"BackendType.is_locked unlocks: {unlocks}")
        else:  # pragma: no cover
            pass

        ret = choose_winner()

        return ret

    def compose(self):
        """subclass must make its own implementation and call,
        :code:`super().compose()`.

        Perform sanity checks
        :raises:

            - :py:exc:`drain_swamp.exceptions.MissingRequirementsFoldersFiles` --
              No requirements folders and files. Abort and provide user feedback

        """
        # Confirm requirements folders and files exist
        gen_files = self.in_files()
        if len(list(gen_files)) == 0:
            msg_exc = "There are no requirements folders and files. Prepare these"
            raise MissingRequirementsFoldersFiles(msg_exc)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_version_file_dump.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_version_file_dump.py

"""

from pathlib import (
    Path,
    PurePath,
)
from unittest.mock import patch

import pytest

from drain_swamp.monkey.wrap_get_version import (
    SEM_VERSION_FALLBACK_SANE,
    scm_version,
    write_to_file,
)
from drain_swamp.version_file.dump_version import (
    _validate_template,
    dump_version,
    write_version_files,
)

from .wd_wrapper import WorkDir


@pytest.fixture()
def wd(wd: WorkDir) -> WorkDir:
    # Create project base folder with project name
    path_new = wd.cwd / "complete_awesome_perfect"
    path_new.mkdir()
    wd.cwd = path_new

    wd("git init")
    wd("git config user.email test@example.com")
    wd('git config user.name "a test"')
    wd.add_command = "git add ."
    wd.commit_command = "git commit --no-verify --no-gpg-sign -m test-{reason}"
    return wd


def test_dump_version(
    tmp_path,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_dump_version" tests
    # prepare
    seq_rel_paths = ("_version.py",)
    prepare_folders_files(seq_rel_paths, tmp_path)

    version = "0.0.1a.dev1"
    # absolute path obsolete --> DeprecationWarning
    write_to = tmp_path / "_version.py"
    with pytest.deprecated_call():
        dump_version(tmp_path, version, write_to)


testdata_write_to_file = (
    (
        Path(__file__).parent.joinpath(
            "_project",
            "install_minimum.pyproject_toml",
        ),
        "complete_awesome_perfect",
        "0.0.2",
    ),
)
ids_write_to_file = ("write version file verify version str",)


@pytest.mark.parametrize(
    "p_toml_file, app_name, kind",
    testdata_write_to_file,
    ids=ids_write_to_file,
)
def test_write_to_file(
    p_toml_file,
    app_name,
    kind,
    wd,
    prep_pyproject_toml,
    prepare_folders_files,
    verify_tag_version,
):
    """Assumes a real Python project skeleton"""
    # pytest --showlocals --log-level INFO -k "test_write_to_file" tests
    assert wd.cwd.exists() and wd.cwd.is_dir()

    # prepare
    #    .gitignore
    msg_gitignore = (
        "dist/*.whl\n"
        "dist/*.tar.gz\n"
        "requirements/*.lnk\n"
        "src/*.egg-info/*\n"
        "src/*/_version.py\n"
    )
    wd.write(".gitignore", msg_gitignore)

    #    pyproject.toml
    p_config = prep_pyproject_toml(p_toml_file, wd.cwd)

    #    empty files
    req_seq = (
        "dist/empty.txt",
        f"src/{app_name}/__init__.py",
        f"src/{app_name}/__main__.py",
        # f"src/{app_name}/_version.py",
        "requirements/prod.in",
        "requirements/pip.in",
        "requirements/manage.in",
        "requirements/prod.unlock",
        "requirements/pip.unlock",
        "requirements/manage.unlock",
        "README.rst",
        "LICENSE",
    )
    prepare_folders_files(req_seq, wd.cwd)
    #   Only wanted the folder
    wd.cwd.joinpath("dist", "empty.txt").unlink()

    #    src/complete_awesome_perfect/__main__.py
    path_dir_src = Path(__file__).parent.joinpath(
        "_project",
        "src",
        app_name,
    )
    path_dir_dest = wd.cwd / "src" / app_name
    path_f = path_dir_src.joinpath("__main__.py")
    prep_pyproject_toml(path_f, path_dir_dest, rename="__main__.py")

    # Act
    #    write to nonexistant version file
    write_to_file(p_config, kind, is_only_not_exists=True)
    #    None --> False
    write_to_file(p_config, kind, is_only_not_exists=None)

    # Verify version file contains "0.0.2"
    assert verify_tag_version(wd.cwd, kind) is True


def test_validate_template(tmp_path, prepare_folders_files):
    # prepare
    target = Path("_version.py")
    assert issubclass(type(target), PurePath)
    seq_rel_paths = (target,)
    prepare_folders_files(seq_rel_paths, tmp_path)

    # supplying a nonsense template
    invalids = (
        "",
        "   ",
    )
    for template in invalids:
        with pytest.warns(Warning):
            actual = _validate_template(target, template)
            assert actual is not None
            assert isinstance(actual, str)
            assert len(actual) != 0

    template = None
    # unsupported target suffix type --> ValueError
    target = Path("_version.rst")
    with pytest.raises(ValueError):
        _validate_template(target, template)

    # supported target suffix
    target = Path("_version.py")
    actual = _validate_template(target, template)
    assert actual is not None
    assert isinstance(actual, str)
    assert len(actual) != 0

    # supply a template
    template_rst = {".rst": "{version}"}
    target = Path("_version.rst")
    actual = _validate_template(target, template_rst)
    assert actual is not None
    assert isinstance(actual, dict)


def test_write_version_files(
    wd,
    prep_pyproject_toml,
    prepare_folders_files,
    verify_tag_version,
):
    # pytest --showlocals --log-level INFO -k "test_write_version_files" tests
    kind = "0.0.2"
    app_name = "complete_awesome_perfect"

    # prepare
    #    pyproject.toml
    p_toml_file = Path(__file__).parent.joinpath(
        "_project",
        "install_minimum.pyproject_toml",
    )
    prep_pyproject_toml(p_toml_file, wd.cwd)

    #    empty files
    req_seq = (
        "dist/empty.txt",
        f"src/{app_name}/__init__.py",
        f"src/{app_name}/__main__.py",
        f"src/{app_name}/_version.py",
    )
    prepare_folders_files(req_seq, wd.cwd)
    #   Only wanted the folder
    wd.cwd.joinpath("dist", "empty.txt").unlink()

    # Act
    #     version file exist, so write is skipped
    p_ver_file = wd.cwd.joinpath("src", app_name, "_version.py")
    assert len(p_ver_file.read_text()) == 0
    write_version_files(
        kind,
        wd.cwd,
        None,
        f"src/{app_name}/_version.py",
        is_only_not_exists=True,
    )
    # Verify
    #     version file exists and is empty. Cuz exists, write skipped
    assert len(p_ver_file.read_text()) == 0

    #     None --> False. Write version file
    write_version_files(
        kind,
        wd.cwd,
        None,
        f"src/{app_name}/_version.py",
        is_only_not_exists=None,
    )
    assert verify_tag_version(wd.cwd, kind) is True


testdata_scm_version_fail = (
    (
        Path(__file__).parent.joinpath(
            "_project",
            "install_minimum.pyproject_toml",
        )
    ),
)
ids_scm_version_fail = ("_get_version returns None --> sane default",)


@pytest.mark.parametrize(
    "p_toml_file",
    testdata_scm_version_fail,
    ids=ids_scm_version_fail,
)
def test_scm_version_fail(
    p_toml_file,
    wd,
    prep_pyproject_toml,
):
    # pytest --showlocals --log-level INFO -k "test_scm_version_fail" tests
    # prepare
    #    pyproject.toml
    path_config = prep_pyproject_toml(p_toml_file, wd.cwd)
    config_abspath = str(path_config)

    valids = (
        (None, SEM_VERSION_FALLBACK_SANE),
        ("0.0.2", "0.0.2"),
    )
    for valid, expected in valids:
        with patch(
            "drain_swamp.monkey.wrap_get_version._get_version",
            return_value=valid,
        ):
            sem_ver_str = scm_version(config_abspath)
            assert sem_ver_str == expected



from collections.abc import Mapping
from typing import Any

from drain_swamp.monkey.hooks import markers

def _is_set_lock(
    config_settings: Mapping[str, Any],
    default: Any | None = False,
) -> bool | None: ...
@markers.hook_impl(tryfirst=True, specname="ds_before_version_infer")  # noqa: Y020
def before_version_infer(config_settings: Mapping[str, Any]) -> str | None: ...



import logging
from functools import cache
from pathlib import Path
from typing import (
    Any,
    Final,
)

_logger: logging.Logger
__all__ = ("PackageMetadata",)

AUTHOR_NAME_FALLBACK: Final[str]

@cache
def get_author_and_email(app_name: str) -> tuple[str, str | None]: ...

class PackageMetadata:
    __slots__ = ("_app_name", "_full_name", "_email", "_d_pyproject_toml")
    def __init__(self, app_name: Any, path: Path | None = None) -> None: ...
    @property
    def app_name(self) -> str: ...
    @app_name.setter
    def app_name(self, val: Any) -> None: ...
    @property
    def full_name(self) -> str: ...
    @property
    def left_name(self) -> str: ...
    @property
    def email(self) -> str | None: ...
    @property
    def d_pyproject_toml(self) -> dict[str, Any] | None: ...



"""
.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_wrap_infer_version.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_wrap_infer_version.py

.. seealso::

   https://github.com/pypa/setuptools_scm/blob/main/testing/test_integration.py
   https://github.com/pypa/setuptools_scm/blob/main/testing/wd_wrapper.py

"""

import logging
import logging.config
import shutil
import sys
from pathlib import Path

import pytest
from setuptools_scm._version_cls import _version_as_tuple

from drain_swamp._run_cmd import run_cmd
from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.monkey.config_settings import ConfigSettings
from drain_swamp.monkey.wrap_infer_version import _rage_quit

from .wd_wrapper import WorkDir


@pytest.fixture()
def wd(wd: WorkDir) -> WorkDir:
    # Create project base folder with project name
    path_new = wd.cwd / "complete_awesome_perfect"
    path_new.mkdir()
    wd.cwd = path_new

    wd("git init")
    wd("git config user.email test@example.com")
    wd('git config user.name "a test"')
    wd.add_command = "git add ."
    wd.commit_command = "git commit --no-verify --no-gpg-sign -m test-{reason}"
    return wd


testdata_dist_get_cmdline_options = (
    (
        Path(__file__).parent.joinpath(
            "_project",
            "install_minimum.pyproject_toml",
        ),
        "complete_awesome_perfect",
        "0.2.0",
        "0.2.1",
    ),
)
ids_dist_get_cmdline_options = (
    "write config_settings .toml file to pass in kwargs to build plugins",
)


@pytest.mark.parametrize(
    "p_toml_file, app_name, commit_version_str, kind",
    testdata_dist_get_cmdline_options,
    ids=ids_dist_get_cmdline_options,
)
def test_dist_get_cmdline_options(
    p_toml_file,
    app_name,
    commit_version_str,
    kind,
    prep_pyproject_toml,
    prepare_folders_files,
    caplog,
    has_logging_occurred,
    wd: WorkDir,
    monkeypatch: pytest.MonkeyPatch,
    verify_tag_version,
):
    """Can access command line options using Distribution.get_cmdline_options

    In tmp path

    .. code-block:: shell

       cd complete_awesome_perfect
       DS_CONFIG_SETTINGS=../setuptools-build.toml ~/Downloads/git_decimals/drain_swamp/.venv/bin/python -m build --no-isolation --sdist

    """
    # pytest --showlocals --log-level INFO -k "test_dist_get_cmdline_options" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    p_base = wd.cwd.parent
    logger.info(f"wd.cwd: {wd.cwd}")
    assert wd.cwd.exists() and wd.cwd.is_dir()

    logger.info(f"sys.executable: {sys.executable}")
    logger.info(f"""scm-version path: {shutil.which("scm-version")}""")

    f_name = f"{app_name}-{kind}.tar.gz"

    # prepare
    #    .gitignore
    msg_gitignore = (
        "dist/*.whl\n"
        "dist/*.tar.gz\n"
        "requirements/*.lnk\n"
        "src/*.egg-info/*\n"
        "src/*/_version.py\n"
        f"{ConfigSettings.FILE_NAME_DEFAULT}\n\n"
    )
    wd.write(".gitignore", msg_gitignore)

    #    empty files
    req_seq = (
        "dist/empty.txt",
        f"src/{app_name}/__init__.py",
        f"src/{app_name}/__main__.py",
        f"src/{app_name}/_version.py",
        "requirements/prod.in",
        "requirements/pip.in",
        "requirements/manage.in",
        "requirements/prod.unlock",
        "requirements/pip.unlock",
        "requirements/manage.unlock",
        "README.rst",
        "LICENSE",
    )
    prepare_folders_files(req_seq, wd.cwd)
    #   Only wanted the folder
    wd.cwd.joinpath("dist", "empty.txt").unlink()

    #    src/complete_awesome_perfect/__main__.py
    path_dir_src = Path(__file__).parent.joinpath(
        "_project",
        "src",
        app_name,
    )
    path_dir_dest = wd.cwd / "src" / app_name
    path_f = path_dir_src.joinpath("__main__.py")
    prep_pyproject_toml(path_f, path_dir_dest, rename="__main__.py")

    #    src/complete_awesome_perfect/_version.py
    str_version_file = (
        f"""__version__ = version = '{commit_version_str}'\n"""
        "__version_tuple__ = version_tuple = "
        f"""{str(_version_as_tuple(commit_version_str))}\n\n"""
    )
    path_f = path_dir_dest.joinpath("_version.py")
    path_f.write_text(str_version_file)

    wd.add_and_commit()
    wd(f"git tag --no-sign -m 'skeleton package' {commit_version_str}")

    # monkeypatch.chdir(p_package_base)

    # export DIST_EXTRA_CONFIG=/tmp/setuptools-build.cfg
    # echo -e '[build_ext]\nparallel = 8\n[bdist_wheel]\npy_limited_api = cp311' > $DIST_EXTRA_CONFIG
    # export DS_CONFIG_SETTINGS=/tmp/setuptools-build.toml
    """cat <<-EOF > "$DS_CONFIG_SETTINGS"
    [project]
    name = "whatever"
    [tool.config-settings]
    kind="0.2.1"
    set-lock="0"
    EOF
    """
    toml_path = str(p_base.joinpath(ConfigSettings.FILE_NAME_DEFAULT))
    toml_contents = (
        "[project]\n"
        f"""name = "{app_name}"\n"""
        """version = "99.99.99a1.dev6"\n"""
        f"[tool.{ConfigSettings.SECTION_NAME}]\n"
        f"""kind="{kind}"\n"""
        """set-lock="0"\n\n"""
    )
    names = (
        ConfigSettings.FILE_NAME_DEFAULT,
        None,
        1.234,
    )
    for file_name in names:
        # modifies os.environ AND writes config settings file
        cs = ConfigSettings(file_name=file_name)
        cs.write(p_base, toml_contents)

    env = None
    assert ConfigSettings.get_abs_path() == toml_path

    cwd = wd.cwd
    cmd = (
        sys.executable,
        "-m",
        "build",
        "--no-isolation",
        "--sdist",
        f"-C--kind='{kind}'",  # unfortunately not passed on by setuptools
        "-C--set-lock='0'",  # unfortunately not passed on by setuptools
    )

    # prepare
    #    pyproject.toml
    prep_pyproject_toml(p_toml_file, wd.cwd)

    # plugins: refresh_links (needs set-lock), cli_scm_version.py (needs kind)
    t_ret = run_cmd(cmd, cwd=cwd, env=env)
    out, err, exit_code, exc = t_ret
    logger.info(f"out {out!r}")
    logger.info(f"err {err!r}")
    logger.info(list(cwd.glob("**/*.tar.gz")))

    assert has_logging_occurred(caplog)
    assert exc is None
    assert exit_code == 0
    path_dist_dir = cwd.joinpath("dist", f_name)
    assert path_dist_dir.exists()

    # logger.info(f"version file contents:\n{path_f.read_text()}")
    assert verify_tag_version(wd.cwd, kind) is True


testdata_infer_version_fail = (
    (
        "0.0.2",
        "0",
        1,
        "does not appear to be a Python project: no pyproject.toml or setup.py",
    ),
)
ids_infer_version_fail = ("missing pyproject.toml. build fails",)


@pytest.mark.parametrize(
    "kind, set_lock, exit_code_expected, msg_fragment_expected",
    testdata_infer_version_fail,
    ids=ids_infer_version_fail,
)
def test_infer_version_fail(
    kind,
    set_lock,
    exit_code_expected,
    msg_fragment_expected,
    wd,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_infer_version_fail" tests

    # missing pyproject.toml build calls infer_version. Which runs build plugins
    cmd = (
        sys.executable,
        "-m",
        "build",
        "--no-isolation",
        "--sdist",
        f"-C--kind='{kind}'",  # unfortunately not passed on by setuptools
        f"-C--set-lock='{set_lock}'",  # unfortunately not passed on by setuptools
    )
    t_ret = run_cmd(cmd, cwd=wd.cwd)
    out, err, exit_code, exc = t_ret
    assert exit_code == exit_code_expected
    assert out.endswith(msg_fragment_expected)

    # _rage_quit
    #    Called by drain_swamp.monkey.wrap_infer_version::infer_version
    #    If 1st arg is None, log a warning and quit
    msg = "Hello World"
    with pytest.raises(SystemExit):
        #    logs a warning
        _rage_quit(None, msg)
    _rage_quit("Not None", msg)


testdata_config_settings_read = (
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            "[tool.incorrect-section-name]\n"
            """kind="current"\n"""
            """set-lock="0"\n\n"""
        ),
        0,
        True,
    ),
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            f"[tool.{ConfigSettings.SECTION_NAME}]\n"
            """kind="current"\n"""
            """set-lock="0"\n\n"""
        ),
        2,
        False,
    ),
    (
        (
            "[project]\n"
            """name = "great-package"\n"""
            """version = "99.99.99a1.dev6"\n"""
            f"[tool.{ConfigSettings.SECTION_NAME}]\n"
        ),
        0,
        True,
    ),
)
ids_config_settings_read = (
    "Incorrect section title. Expecting tool.config-settings",
    "Has section 2 keys",
    "Has section 0 keys",
)


@pytest.mark.parametrize(
    "toml_contents, section_key_count, has_log_warning",
    testdata_config_settings_read,
    ids=ids_config_settings_read,
)
def test_config_settings_read(
    toml_contents,
    section_key_count,
    has_log_warning,
    tmp_path,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_get_config_settings" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    # prepare
    cs = ConfigSettings()
    cs.write(tmp_path, toml_contents)

    path_toml = tmp_path.joinpath(ConfigSettings.FILE_NAME_DEFAULT)
    assert path_toml.exists() and path_toml.is_file()

    ConfigSettings.set_abs_path(path_toml)

    # act
    cs = ConfigSettings()
    d_section = cs.read()
    actual_count = len(d_section.keys())
    assert actual_count == section_key_count

    if has_log_warning:
        assert has_logging_occurred(caplog)


@pytest.mark.parametrize(
    "toml_contents, section_key_count, has_log_warning",
    testdata_config_settings_read,
    ids=ids_config_settings_read,
)
def test_config_settings_set_abs_path(
    toml_contents,
    section_key_count,
    has_log_warning,
    tmp_path,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_get_config_settings" tests
    # prepare
    ConfigSettings.remove_abs_path()

    path_toml = tmp_path.joinpath(ConfigSettings.FILE_NAME_DEFAULT)
    path_toml.write_text(toml_contents)
    assert path_toml.exists() and path_toml.is_file()

    toml_path = str(path_toml)

    invalids = (
        None,
        0.1234,
    )
    for invalid in invalids:
        ConfigSettings.set_abs_path(invalid)
        assert ConfigSettings.get_abs_path() is None

    valids = (path_toml, str(path_toml))
    for valid in valids:
        ConfigSettings.set_abs_path(valid)
        assert ConfigSettings.get_abs_path() == toml_path



import os
from pathlib import Path
from typing import (
    Any,
    Final,
)

__all__ = (
    "seed_changelog",
    "edit_for_release",
    "build_package",
    "pretag",
    "print_cheats",
    "get_tag_version",
    "write_version_file",
)
UNRELEASED: Final[str]
SCRIV_START: Final[str]
COPYRIGHT_START_YEAR_FALLBACK: Final[int]
REGEX_COPYRIGHT_LINE: Final[str]

def update_file(fname: str, pattern: str, replacement: str) -> None: ...
def seed_changelog(path_cwd: Path) -> None: ...
def edit_for_release(
    path_cwd: Path,
    kind: str,
    snippet_co: str | None = None,
) -> int | None: ...
def build_package(
    path: Path,
    kind: str,
) -> bool: ...
def pretag(tag: str) -> tuple[bool, str]: ...
def get_current_version(path: Path) -> str | None: ...
def _get_branch() -> str: ...
def print_cheats(path: Path, kind: str, package_name: str | None = None) -> None: ...

class AlterEnv:
    __slots__ = ("_path_cwd", "_pkg_name", "_scm_override_val", "_scm_override_key")
    def __init__(self, path: Path, kind: str) -> None: ...
    def modify_env(self) -> os._Environ[Any]: ...
    @property
    def path_cwd(self) -> Path: ...
    @property
    def scm_key(self) -> str: ...
    @scm_key.setter
    def scm_key(self, val: Any) -> None: ...
    @property
    def scm_val(self) -> str: ...
    @property
    def pkg_name(self) -> str: ...
    @property
    def version_file(self) -> Path: ...
    @version_file.setter
    def version_file(self, val: Any) -> None: ...

def get_tag_version(path: Path) -> str: ...
def write_version_file(path: Path, kind: str, is_test: bool | None = False) -> None: ...



import logging
from typing import Any

import setuptools
from pluggy import PluginManager

log: logging.Logger

def inspect_pm(pm: PluginManager) -> None: ...
def run_build_plugins(d_config_settings: dict[str, Any]) -> None: ...
def _rage_quit(is_maybe_none: Any | None, msg: str) -> None: ...
def infer_version(dist: setuptools.Distribution) -> None: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Plugin to refresh dependency lock .lnk files. Which are symlinks to respective
.unlock or .lock file

.. seealso::

   `kedro hooks <https://docs.kedro.org/en/latest/extend_kedro/plugins.html#hooks>`_
   `define impl <https://docs.kedro.org/en/latest/hooks/introduction.html#define-the-hook-implementation>`_

"""

from __future__ import annotations

import logging
import traceback
from pathlib import Path
from typing import Any

from drain_swamp.backend_abc import BackendType
from drain_swamp.backend_setuptools import BackendSetupTools  # noqa: F401
from drain_swamp.check_type import click_bool
from drain_swamp.exceptions import (
    BackendNotSupportedError,
    MissingRequirementsFoldersFiles,
    PyProjectTOMLParseError,
    PyProjectTOMLReadError,
)
from drain_swamp.lock_toggle import refresh_links
from drain_swamp.monkey.hooks import markers
from drain_swamp.monkey.hooks.constants import HOOK_NAMESPACE

log = logging.getLogger(__name__)


def _is_set_lock(config_settings, default=None):
    """From config, determine should dependencies be locked or unlocked.

    :param config_settings: config dict
    :type config_settings: dict[str, typing.Any] | None
    :param default: Default None. Create symlink to .unlock file
    :type default: Any | None
    :returns: True lock dependencies otherwise keep the dependencies unlocked
    :rtype: bool | None

    .. seealso::

       `click.BOOL <https://click.palletsprojects.com/en/latest/parameters/#parameter-types>`_
       acceptable true and false strings
       “1”, “true”, “t”, “yes”, “y”, and “on”
       “0”, “false”, “f”, “no”, “n”, and “off”

    """
    is_default_ng = default is not None and not isinstance(default, bool)
    if is_default_ng:
        default = None
    else:  # pragma: no cover
        pass

    is_not_dict = config_settings is None or not isinstance(config_settings, dict)
    if is_not_dict:
        # Default is dependency unlocked
        str_set_lock = None
    else:
        if "--set-lock" in config_settings.keys():
            # Not going to setuptools by call to :code:`python -m build`
            str_set_lock = config_settings["--set-lock"]
        elif "set-lock" in config_settings.keys():
            str_set_lock = config_settings["set-lock"]
        else:
            # Default is dependency unlocked
            str_set_lock = None

    # Simulates click.Bool
    ret = click_bool(val=str_set_lock)
    if ret is None:
        ret = default

    return ret


def _parent_dir(config_settings, default=None):
    """A hack just for testing. Into config_settings, add parent-dir.
    Which is to be feed into BackendType.load_factory call

    :param config_settings: config dict
    :type config_settings: dict[str, typing.Any] | None
    :param default: Default False. Create symlink to .unlock file
    :type default: Any | None
    :returns: parent_dir value to be used with BackendType.load_factory
    :rtype: pathlib.Path
    :meta private
    """
    is_default_ng = default is None or not issubclass(type(default), str)
    if is_default_ng:
        default = None
    else:  # pragma: no cover
        default = Path(default)

    is_not_dict = config_settings is None or not isinstance(config_settings, dict)
    if is_not_dict:
        # Default is dependency unlocked
        ret = None
    else:
        if "--parent-dir" in config_settings.keys():
            # Not going to setuptools by call to :code:`python -m build`
            str_ret = config_settings["--parent-dir"]
            ret = Path(str_ret)
        elif "parent-dir" in config_settings.keys():
            str_ret = config_settings["parent-dir"]
            ret = Path(str_ret)
        else:
            # Default is dependency unlocked
            ret = None

    return ret


@markers.hook_impl(tryfirst=True, specname=f"{HOOK_NAMESPACE}_before_version_infer")
def before_version_infer(config_settings: dict[str, Any]) -> str | None:
    """Refresh .lnk files. Symlink which points to respective
    .unlock or .lock files

    Is not dependent on state. Hook implementation can be run early

    pluggy will inspect this method's signature. So keep it!

    :param config_settings: config dict
    :type config_settings: dict[str, typing.Any]
    :raises:

       - :py:exc:`PyProjectTOMLReadError`
       - :py:exc:`PyProjectTOMLParseError`
       - :py:exc:`BackendNotSupportedError`
       - :py:exc:`MissingRequirementsFoldersFiles`
       - :py:exc:`AssertionError`

    """
    mod_path = "backend plugin ds_refresh_links"
    cwd = Path.cwd()
    # If is_set_lock is None, get lock state from pyproject.toml
    is_set_lock = _is_set_lock(config_settings, default=None)
    parent_dir = _parent_dir(config_settings)

    msg_info = f"{mod_path} is_set_lock: {is_set_lock} parent_dir {parent_dir}"
    log.info(msg_info)

    try:
        inst = BackendType.load_factory(cwd, parent_dir=parent_dir)
        log.info(
            f"{mod_path} inst.path_config: {inst.path_config} "
            f"abs?: {inst.path_config.is_absolute()}"
        )
        refresh_links(inst, is_set_lock=is_set_lock)
    except PyProjectTOMLReadError:
        msg_exc = (
            f"Either not a file or lacks read permissions. {traceback.format_exc()}"
        )
        ret = msg_exc
    except PyProjectTOMLParseError:
        msg_exc = f"Cannot parse pyproject.toml. {traceback.format_exc()}"
        ret = msg_exc
    except BackendNotSupportedError as e:
        """cause typically code neglecting to import backend drivers
        after importing BackendType. Drivers not automatically imported"""
        msg_exc = f"{e} {traceback.format_exc()}"
        ret = msg_exc
    except MissingRequirementsFoldersFiles:
        msg_exc = (
            "Missing requirements folders and files. Prepare these "
            f"{traceback.format_exc()}"
        )
        ret = msg_exc
    except AssertionError:
        msg_exc = (
            "Either no pyproject.toml section, tool.setuptools.dynamic "
            "or no dependencies key"
        )
        ret = msg_exc
    else:
        ret = None

    return ret



import os
from collections.abc import Sequence
from pathlib import Path

__all__ = ("run_cmd",)

def run_cmd(
    cmd: Sequence[str],
    cwd: Path | None = None,
    env: os._Environ[str] | None = None,
) -> tuple[str | None, str | None, int | None, str | None]: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

..

Unittest for module, check_type

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_check_type.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_check_type.py

Needs a config file to specify exact files to include / omit from report.
Will fail with exit code 1 even with 100% coverage

.. seealso::

   https://github.com/pytest-dev/pytest-cov/issues/373#issuecomment-1472861775

"""

from pathlib import Path

import pytest

from drain_swamp.check_type import (
    click_bool,
    is_ok,
    is_relative_required,
)
from drain_swamp.constants import SUFFIX_IN

testdata_is_ok = (
    (None, False),
    ("", False),
    (0.123, False),
    ("    ", False),
    ("Hello World!", True),
)
ids_is_ok = (
    "not str",
    "empty string",
    "not str",
    "contains only whitespace",
    "non-empty string",
)


@pytest.mark.parametrize(
    "mystr, expected",
    testdata_is_ok,
    ids=ids_is_ok,
)
def test_is_ok(mystr, expected):
    # pytest --showlocals --log-level INFO -k "test_is_ok" tests
    actual = is_ok(mystr)
    assert actual == expected


testdata_is_relative_required = (
    pytest.param(None, None, False, marks=pytest.mark.xfail(raises=TypeError)),
    pytest.param(None, 1.12345, False, marks=pytest.mark.xfail(raises=TypeError)),
    pytest.param(None, 1, False, marks=pytest.mark.xfail(raises=TypeError)),
    pytest.param(
        None, "Hello world!", False, marks=pytest.mark.xfail(raises=TypeError)
    ),
    pytest.param(None, (), False, marks=pytest.mark.xfail(raises=ValueError)),
    pytest.param(None, [], False, marks=pytest.mark.xfail(raises=ValueError)),
    pytest.param(
        None, (None, 1.12345, 1), False, marks=pytest.mark.xfail(raises=ValueError)
    ),
    (None, (SUFFIX_IN,), False),
    ("requirements/horse.in", (SUFFIX_IN,), False),
    (Path("requirements/horse.in"), (SUFFIX_IN,), True),
    (Path("requirements/horse.in"), ("in",), True),
    (Path("requirements/horse.in"), (".pip",), False),
    (
        Path("requirements/horse.pip"),
        (SUFFIX_IN,),
        False,
    ),
    (Path("requirements/horse.in"), (".tar", ".gz"), False),
)
ids_is_relative_required = (
    "ext unsupported type None",
    "ext unsupported type float",
    "ext unsupported type int",
    "ext unsupported type str disallowed sequence",
    "ext empty sequence tuple",
    "ext empty sequence list",
    "ext sequence containing only unsupported types",
    "Path | None",
    "must be a Path, got str. matching extensions",
    "Path provided matching extensions",
    "autofix missing period",
    "exts different has .in expect .pip",
    "exts different has .pip expect .in",
    "exts different has .in expect .tar or .gz",
)


@pytest.mark.parametrize(
    "relative_path, exts, expected",
    testdata_is_relative_required,
    ids=ids_is_relative_required,
)
def test_is_relative_required(relative_path, exts, expected):
    # pytest --showlocals --log-level INFO -k "test_is_relative_required" tests
    actual = is_relative_required(path_relative=relative_path, extensions=exts)
    assert actual == expected


testdata_click_bool = (
    (None, None),
    ("George", None),
    ("0", False),
    ("off", False),
    ("1", True),
    ("on", True),
)
ids_click_bool = (
    "None",
    "Unknown str",
    "str number indicating a bool value False",
    "off means False",
    "str number indicating a bool value True",
    "off means True",
)


@pytest.mark.parametrize(
    "val, expected",
    testdata_click_bool,
    ids=ids_click_bool,
)
def test_click_bool(val, expected):
    # pytest --showlocals --log-level INFO -k "test_click_bool" tests
    actual = click_bool(val=val)
    assert actual is expected



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_igor_utils.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_igor_utils.py

Needs a config file to specify exact files to include / omit from report.
Will fail with exit code 1 even with 100% coverage

"""

import contextlib
import io
import logging
import logging.config
import re
import subprocess
from pathlib import Path
from unittest.mock import (
    MagicMock,
    Mock,
    patch,
)

import pytest

from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.igor_utils import (
    SCRIV_START,
    UNRELEASED,
    AlterEnv,
    build_package,
    edit_for_release,
    get_current_version,
    get_tag_version,
    get_version_file_path,
    pretag,
    print_cheats,
    seed_changelog,
    update_file,
    write_version_file,
)
from drain_swamp.version_file.dump_version import write_version_files
from drain_swamp.version_semantic import (
    _is_setuptools_scm,
    _scm_key,
)


def test_update_file(tmp_path, prep_pyproject_toml):
    invalids = (
        None,
        "",
        "   ",
        1.2345,
    )
    pattern = re.escape(SCRIV_START)
    replacement = f"{UNRELEASED}\n\nNothing yet.\n\n\n" + SCRIV_START
    for invalid in invalids:
        out = update_file(invalid, pattern, replacement)
        assert out is None

    # not absolute Path
    rel_path = Path("hi/there")
    out = update_file(rel_path, pattern, replacement)
    assert out is None

    # Need a successful replacement, with differencing file content afterwards
    path_change_rst = Path(__file__).parent.joinpath(
        "_changelog_files",
        "CHANGES-empty.rst",
    )
    path_empty = prep_pyproject_toml(path_change_rst, tmp_path, rename="CHANGES.rst")
    existing_text = path_empty.read_text()
    update_file(
        path_empty,
        pattern,
        replacement,
    )
    actual_text = path_empty.read_text()
    assert existing_text != actual_text
    assert "Nothing yet." in actual_text


def test_seed_changelog(tmp_path, prep_pyproject_toml):
    path_change_rst = Path(__file__).parent.joinpath(
        "_changelog_files",
        "CHANGES-empty.rst",
    )
    path_empty = prep_pyproject_toml(path_change_rst, tmp_path, rename="CHANGES.rst")
    existing_text = path_empty.read_text()
    seed_changelog(tmp_path)
    actual_text = path_empty.read_text()
    assert existing_text != actual_text
    assert "Nothing yet." in actual_text


testdata_edit_for_release = (
    (
        Path(__file__).parent.joinpath("_changelog_files", "CHANGES-empty.rst"),
        Path(__file__).parent.joinpath("_changelog_files", "NOTICE-empty.txt"),
        Path(__file__).parent.joinpath(
            "test_snip", "test_snip_harden_one_snip__with_id_.txt"
        ),
        Path(__file__).parent.joinpath("_good_files", "no_project_name.pyproject_toml"),
        "asdf",
        "0.0.1",
        2,
    ),
    (
        Path(__file__).parent.joinpath("_changelog_files", "CHANGES-empty.rst"),
        Path(__file__).parent.joinpath("_changelog_files", "NOTICE-empty.txt"),
        Path(__file__).parent.joinpath(
            "test_snip", "test_snip_harden_one_snip__with_id_.txt"
        ),
        Path(__file__).parent.joinpath("_good_files", "no_copyright.pyproject_toml"),
        "asdf",
        "0.0.1",
        None,
    ),
    (
        Path(__file__).parent.joinpath("_changelog_files", "CHANGES-empty.rst"),
        Path(__file__).parent.joinpath("_changelog_files", "NOTICE-empty.txt"),
        Path(__file__).parent.joinpath(
            "test_snip", "test_snip_harden_one_snip__with_id_.txt"
        ),
        Path(__file__).parent.joinpath("_good_files", "no_copyright.pyproject_toml"),
        "asdf",
        "0.0.1.dev0",
        None,
    ),
    (
        Path(__file__).parent.joinpath("_changelog_files", "CHANGES-empty.rst"),
        Path(__file__).parent.joinpath("_changelog_files", "NOTICE-empty.txt"),
        Path(__file__).parent.joinpath(
            "test_snip", "test_snip_harden_one_snip__with_id_.txt"
        ),
        Path(__file__).parent.joinpath("_good_files", "weird_copyright.pyproject_toml"),
        "asdf",
        "0.0.1",
        None,
    ),
    (
        Path(__file__).parent.joinpath("_changelog_files", "CHANGES-empty.rst"),
        Path(__file__).parent.joinpath("_changelog_files", "NOTICE-empty.txt"),
        Path(__file__).parent.joinpath(
            "test_snip", "test_snip_harden_one_snip__with_id_.txt"
        ),
        Path(__file__).parent.joinpath(
            "_good_files", "weird_copyright-2.pyproject_toml"
        ),
        "asdf",
        "0.0.1",
        None,
    ),
)
ids_edit_for_release = (
    "has copyright start year",
    "no copyright start year",
    "no copyright start year. dev release",
    "float copyright start year",
    "str not an int copyright start year",
)


@pytest.mark.parametrize(
    "path_changes, path_notice, path_snip, path_pyproject_toml, snippet_co, sem_version_str, expected_exit_co",
    testdata_edit_for_release,
    ids=ids_edit_for_release,
)
def test_edit_for_release(
    path_changes,
    path_notice,
    path_snip,
    path_pyproject_toml,
    snippet_co,
    sem_version_str,
    expected_exit_co,
    tmp_path,
    prep_pyproject_toml,
    caplog,
    has_logging_occurred,
):
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    cmd = []
    kind = sem_version_str

    # prepare
    prep_pyproject_toml(path_changes, tmp_path, rename="CHANGES.rst")
    prep_pyproject_toml(path_notice, tmp_path, rename="NOTICE.txt")

    # before prepare of docs/conf.py
    opt_int = edit_for_release(tmp_path, kind, snippet_co=snippet_co)
    assert opt_int is not None
    assert opt_int == 3

    #    create docs/ folder
    path_docs = tmp_path.joinpath("docs")
    path_docs.mkdir()

    opt_int = edit_for_release(tmp_path, kind, snippet_co=snippet_co)
    assert opt_int is not None
    assert opt_int == 4

    #    docs/conf.py
    prep_pyproject_toml(path_snip, path_docs, rename="conf.py")

    #    pyproject.toml
    prep_pyproject_toml(
        path_pyproject_toml,
        tmp_path,
        rename="pyproject.toml",
    )

    # unittest.mock.patch to prevent changes to _version.py
    # properties; defang so doesn't change src/[proj name]/_version.py
    with (
        patch(
            "subprocess.run",
            return_value=subprocess.CompletedProcess(
                cmd,
                returncode=0,
                stdout=sem_version_str,
            ),
        ),
    ):
        opt_int = edit_for_release(tmp_path, kind, snippet_co=snippet_co)
    if expected_exit_co is None:
        assert opt_int is expected_exit_co
    else:
        assert opt_int == expected_exit_co

    # assert has_logging_occurred(caplog)
    pass


def test_build_package(prep_pyproject_toml, tmp_path):
    """Fake world fake building package"""
    # pytest --showlocals --log-level INFO -k "test_build_package" tests
    # Not a folder
    path_passwd = tmp_path.joinpath("password-secret.txt")
    path_passwd.touch()
    kind = "five element spirit or nature?"
    with pytest.raises(NotADirectoryError):
        build_package(path_passwd, kind)

    # semantic version str is bad --> ValueError
    with pytest.raises(ValueError):
        build_package(tmp_path, kind)

    # package_name is a non-empty str -->
    # build_package --> version_clean --> _tag_version  -> AssertionError
    kind = "tag"
    with (
        patch(
            f"{g_app_name}.version_semantic._tag_version", side_effect=AssertionError
        ),
        pytest.raises(AssertionError),
    ):
        build_package(tmp_path, kind)

    # simulate no initial git tag --> kind either current or now
    sane_fallback = "0.0.1"
    kinds = (
        "now",
        "current",
    )
    for kind in kinds:
        with (
            patch(
                f"{g_app_name}.version_semantic._current_version",
                return_value=sane_fallback,
            ),
            pytest.raises(AssertionError),
        ):
            build_package(tmp_path, kind)

    # simulate no initial git tag --> kind == tag --> sane fallback
    kind = "tag"
    with (
        patch(f"{g_app_name}.version_semantic._tag_version", return_value=None),
        patch(
            f"{g_app_name}.version_semantic._current_version",
            return_value=sane_fallback,
        ),
        pytest.raises(AssertionError),
    ):
        build_package(tmp_path, kind)

    # No preparation of pyproject.toml
    cmd = []
    kind = "0.0.1"
    with (
        patch(
            "subprocess.run",
            side_effect=subprocess.CalledProcessError(128, cmd),
        ),
        pytest.raises(AssertionError),
    ):
        build_package(tmp_path, kind)

    # prepare pyproject.toml
    p_toml_file = Path(__file__).parent.joinpath(
        "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
    )
    prep_pyproject_toml(p_toml_file, tmp_path)

    cmd = []
    kind = "0.0.1"
    with (
        patch(
            "subprocess.run",
            side_effect=subprocess.CalledProcessError(128, cmd),
        ),
    ):
        bool_out = build_package(tmp_path, kind)
        assert bool_out is False

    with (
        patch(
            "subprocess.run",
            return_value=subprocess.CompletedProcess(
                cmd,
                returncode=128,
                stdout="fatal: No names found, cannot describe anything.",
            ),
        ),
    ):
        bool_out = build_package(tmp_path, kind)
        assert bool_out is False

    # build success
    with (
        patch(
            "subprocess.run",
            return_value=subprocess.CompletedProcess(
                cmd,
                returncode=0,
                stdout="You just installed the best package and dependencies ... ever",
            ),
        ),
    ):
        bool_out = build_package(tmp_path, kind)
        assert bool_out is True


testdata_pretag = (
    ("1!v1.0.1+g4b33a80.d20240129", "1.0.1"),
    ("0.1.1.candidate1dev1+g4b33a80.d20240129", "0.1.1rc1.dev1"),
)
ids_pretag = (
    "with epoch locals and prepended v",
    "malformed semantic ver str raise ValueError",
)


@pytest.mark.parametrize(
    "ver, expected",
    testdata_pretag,
    ids=ids_pretag,
)
def test_pretag(ver, expected):
    """Sanitize ver str

    From just this unittest non-obvious how to sanitize a ver str containing an epoch.
    Shell escape doesn't work. Instead surround by single quotes"""
    # pytest --showlocals --log-level INFO -k "test_pretag" tests
    is_success, actual = pretag(ver)
    if is_success:
        assert actual == expected
    else:
        # prints the error message
        assert len(actual) != 0
        # not the fixed semantic version str
        assert actual != expected


def test_get_current_version(path_project_base):
    """Get the current version of this package. Requires package setuptools-scm"""
    ver = get_current_version(path_project_base())
    if _is_setuptools_scm():
        assert isinstance(ver, str)
    else:
        assert ver is None


def test_print_cheats(path_project_base, prep_pyproject_toml, tmp_path):
    """prints non-vital info. Useful urls and commands"""
    # pytest --showlocals --log-level INFO -k "test_print_cheats" tests
    path_cwd = path_project_base()
    kinds = ("tag", "current")

    # NotADirectoryError
    kind = "current"
    path_f = tmp_path.joinpath("fish.txt")
    path_f.touch()
    with pytest.raises(NotADirectoryError):
        print_cheats(path_f, kind)

    # ValueError
    with pytest.raises(ValueError):
        kind_food_bad = "'dog food tastes better than this'"
        print_cheats(path_cwd, kind_food_bad)

    # kind="tag". Missing ``pyproject.toml``. Need to prepare that
    # Only for missing or unparsable results in AssertionError.
    # For all other issues, issues warning and fallsback to current version
    kind = "tag"
    with pytest.raises(AssertionError):
        print_cheats(tmp_path, kind)

    # prepare
    p_toml_file = Path(__file__).parent.joinpath(
        "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
    )
    prep_pyproject_toml(p_toml_file, tmp_path)

    for kind in kinds:
        with contextlib.redirect_stdout(io.StringIO()) as f:
            print_cheats(path_cwd, kind)
        out = f.getvalue()
        assert out is not None
        assert isinstance(out, str)
        assert len(out.strip()) != 0

        branches = (
            "important_branch",
            "master",
            "main",
        )
        for branch in branches:
            with (
                patch(
                    f"{g_app_name}.igor_utils._get_branch",
                    return_value=branch,
                ),
                contextlib.redirect_stdout(io.StringIO()) as f,
            ):
                print_cheats(path_cwd, kind)
            out = f.getvalue()
            assert out is not None
            assert isinstance(out, str)
            assert len(out.strip()) != 0


testdata_get_version_file_path = (
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "no_project_name.pyproject_toml",
        ),
        None,
    ),
)
ids_get_version_file_path = (
    "In pyproject.toml, missing tool.pipenv-unlock.version_file",
)


@pytest.mark.parametrize(
    "p_toml_file, expected_version_file",
    testdata_get_version_file_path,
    ids=ids_get_version_file_path,
)
def test_get_version_file_path(
    p_toml_file,
    expected_version_file,
    tmp_path,
    prep_pyproject_toml,
):
    # pytest --showlocals --log-level INFO -k "test_get_version_file_path" tests

    path_f = tmp_path / "pyproject.toml"
    actual_version_file = get_version_file_path(path_f)
    assert actual_version_file is None

    # prepare
    path_f = prep_pyproject_toml(p_toml_file, tmp_path)

    actual_version_file = get_version_file_path(path_f)
    assert actual_version_file == expected_version_file


testdata_alter_env = (
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "complete-manage-pip-prod-unlock.pyproject_toml",
        ),
        "complete-awesome-perfect",
    ),
)
ids_alter_env = ("Buildable package pyproject.toml",)


@pytest.mark.parametrize(
    "p_toml_file, pkg_name_expected",
    testdata_alter_env,
    ids=ids_alter_env,
)
def test_alter_env(
    p_toml_file,
    pkg_name_expected,
    path_project_base,
    prep_pyproject_toml,
    prepare_folders_files,
    tmp_path,
):
    # pytest --showlocals --log-level INFO -k "test_alter_env" tests
    path_cwd = path_project_base()
    # NotADirectoryError
    kind = "tag"
    path_f = tmp_path.joinpath("fish.txt")
    path_f.touch()
    with pytest.raises(NotADirectoryError):
        AlterEnv(path_f, kind)

    # ValueError
    kind_food_bad = "'dog food tastes better than this'"
    with pytest.raises(ValueError):
        AlterEnv(path_cwd, kind_food_bad)

    # AssertionError -- no pyproject.toml or no project.name field
    kind = "tag"
    with pytest.raises(AssertionError):
        AlterEnv(tmp_path, kind)

    # prepare
    prep_pyproject_toml(p_toml_file, tmp_path)

    """in the tmp folder, there is no git. kind: current returns a fallback"""
    kinds = ("0.0.1",)
    for kind in kinds:
        with (
            patch(
                f"{g_app_name}.igor_utils.SemVersion.version_clean",
                return_value=kind,
            ),
        ):
            ae = AlterEnv(tmp_path, kind)
            assert ae.path_cwd == tmp_path
            assert ae.pkg_name == pkg_name_expected
            assert ae.scm_key == _scm_key(pkg_name_expected)
            assert ae.scm_val is not None
            assert ae.version_file.endswith("_version.py")

            # property setter -- AlterEnv.version_file
            invalids = (0.1, None)
            for invalid in invalids:
                with pytest.raises(TypeError):
                    ae.version_file = invalid

            # prepare -- version file (empty w/ parent folders)
            ver_file = ae.version_file
            p_ver = tmp_path.joinpath(ver_file)
            prepare_folders_files((p_ver,), tmp_path)

            write_version_file(tmp_path, kind, is_test=True)
            # confirm file size is not 0
            assert p_ver.stat().st_size != 0

            # No version file template for file type
            # write_version_file with is_test=True does not validate has template
            ver_file = "src/complete_awesome_perfect/_version.rst"
            ae.version_file = ver_file
            p_ver = ae.path_cwd.joinpath(ver_file)
            prepare_folders_files((p_ver,), tmp_path)

            write_version_file(tmp_path, kind, is_test=True)
            assert p_ver.stat().st_size == 0

            # Simulate missing tool.pipenv-unlock.version_file or non-str
            with (
                patch(
                    f"{g_app_name}.igor_utils.get_version_file_path",
                    return_value=None,
                ),
                pytest.raises(AssertionError),
            ):
                ae = AlterEnv(tmp_path, kind)


def test_write_version_file(tmp_path, prep_pyproject_toml):
    # pytest --showlocals --log-level INFO -k "test_write_version_file" tests
    # prepare
    p_toml_file = Path(__file__).parent.joinpath(
        "_good_files",
        "complete-manage-pip-prod-unlock.pyproject_toml",
    )
    prep_pyproject_toml(p_toml_file, tmp_path)

    kinds = (
        (
            "0.0.1",
            tmp_path / "_version.rst",  # no .rst template
        ),
        (
            "golf balls",  # kind bad
            tmp_path.joinpath("src", "complete_awesome_perfect", "_version.py"),
        ),
    )
    for kind, path_f in kinds:
        mock_cls = MagicMock(wraps=AlterEnv, spec=AlterEnv)
        mock_cls.path_cwd.return_value = tmp_path
        mock_cls.scm_val = Mock(spec=kind, return_value=kind)
        mock_cls.version_file.return_value = str(path_f)

        with (
            patch(
                "drain_swamp.igor_utils.AlterEnv",
                return_value=mock_cls,
            ),
            pytest.raises(ValueError),
        ):
            write_version_file(tmp_path, kind, is_test=True)

    # defang

    with patch("drain_swamp.igor_utils.write_version_files", return_value=None):
        # is_test default --> False
        kind = ("0.0.1",)
        invalids = (
            "Hi there",
            0.1234,
        )
        for invalid in invalids:
            write_version_file(tmp_path, kind, is_test=invalid)

        kind = ("golf balls",)
        with pytest.raises(ValueError):
            write_version_file(tmp_path, kind, is_test=True)


def test_get_tag_version(tmp_path, prep_pyproject_toml, prepare_folders_files):
    # pytest --showlocals --log-level INFO -k "test_get_tag_version" tests
    path_cwd = tmp_path / "complete_awesome_perfect"
    path_cwd.mkdir()
    path_dir_dest = path_cwd.joinpath("src", "complete_awesome_perfect")
    path_version_file = path_dir_dest.joinpath("_version.py")
    version_file_relpath = Path("src").joinpath(
        "complete_awesome_perfect",
        "_version.py",
    )

    # prepare
    srcs = (path_version_file,)
    prepare_folders_files(srcs, path_cwd)

    sem_ver = "0.0.2"
    write_version_files(sem_ver, path_cwd, version_file_relpath, None)

    #    no pyproject.toml
    with (pytest.raises(AssertionError),):
        get_tag_version(path_cwd)

    p_toml_src = Path(__file__).parent.joinpath(
        "_good_files",
        "complete-manage-pip-prod-unlock.pyproject_toml",
    )
    prep_pyproject_toml(p_toml_src, path_cwd)

    #    success
    actual = get_tag_version(path_cwd)
    assert actual == sem_ver

    #    NotADirectory
    path_f = path_cwd.joinpath("fish.txt")
    path_f.touch()
    with pytest.raises(NotADirectoryError):
        get_tag_version(path_f)

    str_version_file = (
        """__version__ = version = 'golf balls'\n"""
        """__version_tuple__ = version_tuple = (0, 0, 3, 'golf balls', '+golfballs.d20241212')\n\n"""
    )
    path_version_file.write_text(str_version_file)

    sane_fallback = "0.0.1"
    sem_ver = get_tag_version(path_cwd)
    assert sem_ver == sane_fallback



import logging
from pathlib import Path
from typing import Final

_logger: logging.Logger
entrypoint_name: Final[str]
help_path: Final[str]
help_snippet_co: Final[str]
help_kind: Final[str]
EPILOG_SEED: Final[str]
EPILOG_EDITS: Final[str]

def main() -> None: ...
def seed(path: Path) -> None: ...
def edit(path: Path, kind: str, snippet_co: str | None = None) -> None: ...
def semantic_version_aware_build(
    path: Path, kind: str, package_name: str | None = None
) -> None: ...
def validate_tag(ver: str) -> None: ...
def current_version(path: Path) -> None: ...
def do_cheats(path: Path, kind: str) -> None: ...
